"""
A bunch of fake API responses for testing.

Because the library has a habit of mangling the data (_process_dict), these are functions that will always return clean responses.
"""

import random
import typing
import discord_typings
from typing import Optional

__all__ = ("SAMPLE_DM_DATA", "SAMPLE_GUILD_DATA", "SAMPLE_USER_DATA")


def SAMPLE_USER_DATA(user_id: Optional[str] = None) -> discord_typings.UserData:
    if user_id is None:
        user_id = "123456789012345678"

    return {
        "id": user_id,
        "username": "test_user",
        "discriminator": "1234",
        "avatar": "",
    }


def SAMPLE_DM_DATA() -> discord_typings.DMChannelData:
    return {
        "id": "123456789012345679",
        "type": 1,
        "last_message_id": None,
        "recipients": [SAMPLE_USER_DATA()],
    }


def SAMPLE_CHANNEL_DATA(
    channel_id: Optional[str] = None, guild_id: Optional[str] = None
) -> discord_typings.ChannelData:
    if channel_id is None:
        channel_id = "123456789012345678"
    return {
        "id": channel_id,
        "type": 0,
        "guild_id": guild_id,
        "name": "test_channel",
        "topic": "",
        "position": 0,
        "permission_overwrites": [],
        "bitrate": 0,
        "user_limit": 0,
        "rate_limit_per_user": 0,
        "last_message_id": None,
        "permissions": 0,
        "nsfw": False,
    }


def SAMPLE_GUILD_DATA(guild_id: Optional[str] = None) -> discord_typings.GuildData:
    if guild_id is None:
        guild_id = "123456789012345670"
    return {
        "id": guild_id,
        "name": "test_guild",
        "icon": "",
        "splash": "",
        "discovery_splash": "",
        "owner_id": "123456789012345678",
        "afk_channel_id": None,
        "afk_timeout": 0,
        "verification_level": 0,
        "default_message_notifications": 0,
        "explicit_content_filter": 0,
        "roles": [],
        "emojis": [],
        "features": [],
        "mfa_level": 0,
        "application_id": None,
        "system_channel_id": None,
        "system_channel_flags": 0,
        "rules_channel_id": None,
        "vanity_url_code": None,
        "description": None,
        "banner": None,
        "premium_tier": 0,
        "preferred_locale": "en-US",
        "public_updates_channel_id": None,
        "nsfw_level": 0,
        "stickers": [],
        "premium_progress_bar_enabled": False,
    }


def SAMPLE_MESSAGE_DATA(
    channel_id: Optional[str] = None,
    user_id: Optional[str] = None,
    message_id: Optional[str] = None,
    guild_id: Optional[str] = None,
) -> discord_typings.MessageCreateData:
    if channel_id is None:
        channel_id = "123456789012345678"
    if message_id is None:
        message_id = "123456789012345677"
    data = {
        "id": message_id,
        "channel_id": channel_id,
        "author": SAMPLE_USER_DATA(user_id),
        "content": "test_message",
        "timestamp": "2022-07-16T20:56:55.999419+01:00",
        "edited_timestamp": None,
        "tts": False,
        "mention_everyone": False,
        "mentions": [SAMPLE_USER_DATA(user_id)],
        "mention_roles": [],
        "mention_channels": [],
        "attachments": [],
        "embeds": [],
        "reactions": [],
        "nonce": None,
        "pinned": False,
        "webhook_id": None,
        "type": 0,
        "activity": None,
        "application": None,
        "application_id": None,
        "message_reference": None,
        "flags": 0,
        "refereces_message": None,
        # "interaction": None,
        "thread": None,
        "components": [],
        "sticker_items": [],
    }
    if guild_id is not None:
        data["guild_id"] = guild_id
    return data


def SAMPLE_LOCALE() -> str:
    options = typing.get_args(discord_typings.Locales)
    return random.sample(options, 1)[0]  # type: ignore


def SAMPLE_APPLICATION_DATA(owner: typing.Optional[discord_typings.UserData]) -> discord_typings.ApplicationData:
    return {
        "id": "10001000010001",
        "name": "Sample Application",
        "description": "Mock data, this is not a real bot.",
        "rpc_origins": [],
        "bot_public": False,
        "bot_require_code_grant": False,
        "owner": owner,
        "verify_key": "",
        "summary": "",
        "icon": None,
        "team": None,
    }



from typing import TYPE_CHECKING

import interactions.api.events as events
from interactions.models import PartialEmoji, Reaction

from ._template import EventMixinTemplate, Processor

if TYPE_CHECKING:
    from interactions.api.events import RawGatewayEvent

__all__ = ("ReactionEvents",)


class ReactionEvents(EventMixinTemplate):
    async def _handle_message_reaction_change(self, event: "RawGatewayEvent", add: bool) -> None:
        if member := event.data.get("member"):
            author = self.cache.place_member_data(event.data.get("guild_id"), member)
        elif guild_id := event.data.get("guild_id"):
            author = await self.cache.fetch_member(guild_id, event.data.get("user_id"))
        else:
            author = await self.cache.fetch_user(event.data.get("user_id"))

        emoji = PartialEmoji.from_dict(event.data.get("emoji"))  # type: ignore
        message = self.cache.get_message(event.data.get("channel_id"), event.data.get("message_id"))
        reaction = None

        if message:
            for i in range(len(message.reactions)):
                r = message.reactions[i]
                if r.emoji == emoji:
                    if add:
                        r.count += 1
                    else:
                        r.count -= 1

                    if r.count <= 0:
                        message.reactions.pop(i)
                    else:
                        message.reactions[i] = r
                    reaction = r
                    break
            else:
                reaction = Reaction.from_dict(
                    {
                        "count": 1,
                        "me": author.id == self.user.id,  # type: ignore
                        "emoji": emoji.to_dict(),
                        "message_id": message.id,
                        "channel_id": message._channel_id,
                    },
                    self,  # type: ignore
                )
                message.reactions.append(reaction)

        else:
            message = await self.cache.fetch_message(event.data.get("channel_id"), event.data.get("message_id"))
            for r in message.reactions:
                if r.emoji == emoji:
                    reaction = r
                    break
        if add:
            self.dispatch(events.MessageReactionAdd(message=message, emoji=emoji, author=author, reaction=reaction))
        else:
            self.dispatch(events.MessageReactionRemove(message=message, emoji=emoji, author=author, reaction=reaction))

    @Processor.define()
    async def _on_raw_message_reaction_add(self, event: "RawGatewayEvent") -> None:
        await self._handle_message_reaction_change(event, add=True)

    @Processor.define()
    async def _on_raw_message_reaction_remove(self, event: "RawGatewayEvent") -> None:
        await self._handle_message_reaction_change(event, add=False)

    @Processor.define()
    async def _on_raw_message_reaction_remove_all(self, event: "RawGatewayEvent") -> None:
        if message := self.cache.get_message(event.data["channel_id"], event.data["message_id"]):
            message.reactions = []
        self.dispatch(
            events.MessageReactionRemoveAll(
                event.data.get("guild_id"),
                await self.cache.fetch_message(event.data["channel_id"], event.data["message_id"]),
            )
        )

    @Processor.define()
    async def _on_raw_message_reaction_remove_emoji(self, event: "RawGatewayEvent") -> None:
        emoji = PartialEmoji.from_dict(event.data.get("emoji"))
        message = self.cache.get_message(event.data.get("channel_id"), event.data.get("message_id"))

        if message:
            for i, reaction in enumerate(message.reactions):
                if reaction.emoji == emoji:
                    message.reactions.pop(i)
                    break
        else:
            message = await self.cache.fetch_message(event.data.get("channel_id"), event.data.get("message_id"))

        self.dispatch(
            events.MessageReactionRemoveEmoji(
                event.data.get("guild_id"),
                message,
                emoji,
            )
        )



from .const import (
    __version__,
    __repo_url__,
    __py_version__,
    __api_version__,
    get_logger,
    logger_name,
    kwarg_spam,
    DISCORD_EPOCH,
    ACTION_ROW_MAX_ITEMS,
    SELECTS_MAX_OPTIONS,
    SELECT_MAX_NAME_LENGTH,
    CONTEXT_MENU_NAME_LENGTH,
    SLASH_CMD_NAME_LENGTH,
    SLASH_CMD_MAX_DESC_LENGTH,
    SLASH_CMD_MAX_OPTIONS,
    SLASH_OPTION_NAME_LENGTH,
    EMBED_MAX_NAME_LENGTH,
    EMBED_MAX_DESC_LENGTH,
    EMBED_MAX_FIELDS,
    EMBED_TOTAL_MAX,
    EMBED_FIELD_VALUE_LENGTH,
    Singleton,
    Sentinel,
    GlobalScope,
    Missing,
    MentionPrefix,
    GLOBAL_SCOPE,
    MISSING,
    MENTION_PREFIX,
    PREMIUM_GUILD_LIMITS,
    POLL_MAX_ANSWERS,
    POLL_MAX_DURATION_HOURS,
    Absent,
    T,
    T_co,
    ClientT,
)
from .client import Client
from .auto_shard_client import AutoShardedClient
from . import smart_cache
from . import errors
from . import utils

__all__ = (
    "__version__",
    "__repo_url__",
    "__py_version__",
    "__api_version__",
    "get_logger",
    "logger_name",
    "kwarg_spam",
    "DISCORD_EPOCH",
    "ACTION_ROW_MAX_ITEMS",
    "SELECTS_MAX_OPTIONS",
    "SELECT_MAX_NAME_LENGTH",
    "CONTEXT_MENU_NAME_LENGTH",
    "SLASH_CMD_NAME_LENGTH",
    "SLASH_CMD_MAX_DESC_LENGTH",
    "SLASH_CMD_MAX_OPTIONS",
    "SLASH_OPTION_NAME_LENGTH",
    "EMBED_MAX_NAME_LENGTH",
    "EMBED_MAX_DESC_LENGTH",
    "EMBED_MAX_FIELDS",
    "EMBED_TOTAL_MAX",
    "EMBED_FIELD_VALUE_LENGTH",
    "POLL_MAX_ANSWERS",
    "POLL_MAX_DURATION_HOURS",
    "Singleton",
    "Sentinel",
    "GlobalScope",
    "Missing",
    "MentionPrefix",
    "GLOBAL_SCOPE",
    "MISSING",
    "MENTION_PREFIX",
    "PREMIUM_GUILD_LIMITS",
    "Absent",
    "T",
    "T_co",
    "ClientT",
    "Client",
    "AutoShardedClient",
    "smart_cache",
    "errors",
    "utils",
)



from typing import Optional, Union, Dict, Any
from typing_extensions import Self

import attrs

from interactions.client.const import MISSING, POLL_MAX_DURATION_HOURS, POLL_MAX_ANSWERS
from interactions.client.utils.attr_converters import (
    optional,
    timestamp_converter,
)
from interactions.client.mixins.serialization import DictSerializationMixin
from interactions.client.utils.serializer import no_export_meta
from interactions.models.discord.emoji import PartialEmoji, process_emoji
from interactions.models.discord.enums import PollLayoutType
from interactions.models.discord.timestamp import Timestamp

__all__ = (
    "PollMedia",
    "PollAnswer",
    "PollAnswerCount",
    "PollResults",
    "Poll",
)


@attrs.define(eq=False, order=False, hash=False, kw_only=True)
class PollMedia(DictSerializationMixin):
    text: Optional[str] = attrs.field(repr=False, default=None)
    """
    The text of the field.

    !!! warning
        While `text` is *marked* as optional, it is *currently required* by Discord's API to make polls.
        According to Discord, this may change to be actually optional in the future.
    """
    emoji: Optional[PartialEmoji] = attrs.field(repr=False, default=None, converter=optional(PartialEmoji.from_dict))
    """The emoji of the field."""

    @classmethod
    def create(cls, *, text: Optional[str] = None, emoji: Optional[Union[PartialEmoji, dict, str]] = None) -> Self:
        """
        Create a PollMedia object, used for questions and answers for polls.

        !!! warning
            While `text` is *marked* as optional, it is *currently required* by Discord's API to make polls.
            According to Discord, this may change to be actually optional in the future.

        Args:
            text: The text of the field.
            emoji: The emoji of the field.

        Returns:
            A PollMedia object.

        """
        if not text and not emoji:
            raise ValueError("Either text or emoji must be provided.")

        return cls(text=text, emoji=process_emoji(emoji))


@attrs.define(eq=False, order=False, hash=False, kw_only=True)
class PollAnswer(DictSerializationMixin):
    poll_media: PollMedia = attrs.field(repr=False, converter=PollMedia.from_dict)
    """The data of the answer."""
    answer_id: Optional[int] = attrs.field(repr=False, default=None)
    """The ID of the answer. This is only returned for polls that have been given by Discord's API."""


@attrs.define(eq=False, order=False, hash=False, kw_only=True)
class PollAnswerCount(DictSerializationMixin):
    id: int = attrs.field(repr=False)
    """The answer ID of the answer."""
    count: int = attrs.field(repr=False, default=0)
    """The number of votes for this answer."""
    me_voted: bool = attrs.field(repr=False, default=False)
    """Whether the current user voted for this answer."""


@attrs.define(eq=False, order=False, hash=False, kw_only=True)
class PollResults(DictSerializationMixin):
    is_finalized: bool = attrs.field(repr=False, default=False)
    """Whether the votes have been precisely counted."""
    answer_counts: list[PollAnswerCount] = attrs.field(repr=False, factory=list, converter=PollAnswerCount.from_list)
    """The counts for each answer."""


@attrs.define(eq=False, order=False, hash=False, kw_only=True)
class Poll(DictSerializationMixin):
    question: PollMedia = attrs.field(repr=False)
    """The question of the poll. Only text media is supported."""
    answers: list[PollAnswer] = attrs.field(repr=False, factory=list, converter=PollAnswer.from_list)
    """Each of the answers available in the poll, up to 10."""
    expiry: Timestamp = attrs.field(repr=False, default=MISSING, converter=optional(timestamp_converter))
    """Number of hours the poll is open for, up to 7 days."""
    allow_multiselect: bool = attrs.field(repr=False, default=False, metadata=no_export_meta)
    """Whether a user can select multiple answers."""
    layout_type: PollLayoutType = attrs.field(repr=False, default=PollLayoutType.DEFAULT, converter=PollLayoutType)
    """The layout type of the poll."""
    results: Optional[PollResults] = attrs.field(repr=False, default=None, converter=optional(PollResults.from_dict))
    """The results of the poll, if the polls is finished."""

    _duration: int = attrs.field(repr=False, default=0)
    """How long, in hours, the poll will be open for (up to 7 days). This is only used when creating polls."""

    @classmethod
    def create(
        cls,
        question: str,
        *,
        duration: int,
        allow_multiselect: bool = False,
        answers: Optional[list[PollMedia | str]] = None,
    ) -> Self:
        """
        Create a Poll object for sending.

        Args:
            question: The question of the poll.
            duration: How long, in hours, the poll will be open for (up to 7 days).
            allow_multiselect: Whether a user can select multiple answers.
            answers: Each of the answers available in the poll, up to 10.

        Returns:
            A Poll object.

        """
        if answers:
            media_to_answers = [
                (
                    PollAnswer(poll_media=answer)
                    if isinstance(answer, PollMedia)
                    else PollAnswer(poll_media=PollMedia.create(text=answer))
                )
                for answer in answers
            ]
        else:
            media_to_answers = []

        return cls(
            question=PollMedia(text=question),
            duration=duration,
            allow_multiselect=allow_multiselect,
            answers=media_to_answers,
        )

    @answers.validator
    def _answers_validation(self, attribute: str, value: Any) -> None:
        if len(value) > POLL_MAX_ANSWERS:
            raise ValueError(f"A poll can have at most {POLL_MAX_ANSWERS} answers.")

    @_duration.validator
    def _duration_validation(self, attribute: str, value: int) -> None:
        if value < 0 or value > POLL_MAX_DURATION_HOURS:
            raise ValueError(
                f"The duration must be between 0 and {POLL_MAX_DURATION_HOURS} hours ({POLL_MAX_DURATION_HOURS // 24} days)."
            )

    def add_answer(self, text: Optional[str] = None, emoji: Optional[Union[PartialEmoji, dict, str]] = None) -> Self:
        """
        Adds an answer to the poll.

        !!! warning
            While `text` is *marked* as optional, it is *currently required* by Discord's API to make polls.
            According to Discord, this may change to be actually optional in the future.

        Args:
            text: The text of the answer.
            emoji: The emoji for the answer.

        """
        if not text and not emoji:
            raise ValueError("Either text or emoji must be provided")

        self.answers.append(PollAnswer(poll_media=PollMedia.create(text=text, emoji=emoji)))
        self._answers_validation("answers", self.answers)
        return self

    def to_dict(self) -> Dict[str, Any]:
        data = super().to_dict()

        data["duration"] = self._duration
        data.pop("_duration", None)
        return data



import datetime
from typing import Union, List, SupportsInt, Optional

import attrs

import interactions.models as models
from interactions.client.const import MISSING, Absent, DISCORD_EPOCH

__all__ = (
    "to_snowflake",
    "to_optional_snowflake",
    "to_snowflake_list",
    "Snowflake",
    "SnowflakeObject",
    "Snowflake_Type",
)


# Snowflake_Type should be used in FUNCTION args of user-facing APIs (combined with to_snowflake to sanitize input)
# For MODEL id fields, just use int as type-hinting instead;
# For attr convertors: Use int() when API-facing conversion is expected,
# use to_snowflake when user should create this object
Snowflake_Type = Union[int, str, "SnowflakeObject", SupportsInt, "Snowflake"]


def to_snowflake(snowflake: Snowflake_Type) -> "Snowflake":
    """
    Helper function to convert something into correct Discord snowflake int, gives more helpful errors Use internally to sanitize user input or in user- facing APIs (a must).

    For Discord-API - facing code, just int() is sufficient

    """
    try:
        snowflake = Snowflake(snowflake)
    except TypeError as e:
        raise TypeError(
            f"ID (snowflake) should be instance of int, str, SnowflakeObject, or support __int__. "
            f"Got '{snowflake}' ({type(snowflake)}) instead."
        ) from e
    except ValueError as e:
        raise ValueError(f"ID (snowflake) should represent int. Got '{snowflake}' ({type(snowflake)}) instead.") from e

    if 22 > snowflake.bit_length() > 64:
        raise ValueError(
            f"ID (snowflake) is not in correct Discord format! Bit length of int should be from 22 to 64 "
            f"Got '{snowflake}' (bit length {snowflake.bit_length()})"
        )

    return snowflake


def to_optional_snowflake(
    snowflake: Absent[Optional[Snowflake_Type]] = MISSING,
) -> "Optional[Snowflake]":
    if snowflake is MISSING:
        return MISSING
    return None if snowflake is None else to_snowflake(snowflake)


def to_snowflake_list(snowflakes: List[Snowflake_Type]) -> "list[Snowflake]":
    return [to_snowflake(c) for c in snowflakes]


class Snowflake(int):
    def __new__(cls, id: int) -> "Snowflake":
        return int.__new__(cls, id)

    def __iadd__(self, other) -> "Snowflake":
        return Snowflake(int(self) + other)

    def __isub__(self, other) -> "Snowflake":
        return Snowflake(int(self) - other)

    def __add__(self, other) -> "Snowflake":
        return Snowflake(int(self) + other)

    def __sub__(self, other) -> "Snowflake":
        return Snowflake(int(self) - other)

    @classmethod
    def from_datetime(cls, dt: datetime.datetime, *, high: bool = False) -> "Snowflake":
        """
        Creates a snowflake from a datetime object.

        Args:
            dt: The datetime object to create the snowflake from.
            high: Set to True if you're creating a snowflake in the discord future.

        """
        timestamp = dt.timestamp()
        ms = int(timestamp * 1000 - DISCORD_EPOCH)

        return cls((ms << 22) + (2**22 - 1 if high else 0))

    @property
    def created_at(self) -> "models.Timestamp":
        """
        Returns a timestamp representing the date-time this discord object was created.

        :Returns:
        """
        from interactions.models import (
            Timestamp,
        )  # dirty i know; but it's an unavoidable circular import

        return Timestamp.from_snowflake(self)

    @property
    def worker_id(self) -> int:
        """The internal worker ID of the snowflake."""
        return (int(self) & 0x3E0000) >> 17

    @property
    def process_id(self) -> int:
        """The internal process ID of the snowflake."""
        return (int(self) & 0x1F000) >> 12

    @property
    def increment(self) -> int:
        """
        The internal incrementation number of the snowflake.

        This value will only increment when a process has been
        generated on this snowflake, e.g. a resource.
        """
        return int(self) & 0xFFF

    def difference(self, other: Union["Snowflake", int, str]) -> datetime.timedelta:
        """
        Returns the difference between two snowflakes.

        Args:
            other: The other snowflake to compare to.

        """
        if not isinstance(other, Snowflake):
            other = Snowflake(other)
        return abs(self.created_at - other.created_at)


@attrs.define(eq=False, order=False, hash=False, slots=False)
class SnowflakeObject:
    id: Snowflake = attrs.field(repr=True, converter=Snowflake, metadata={"docs": "Discord unique snowflake ID"})

    def __eq__(self, other: "SnowflakeObject") -> bool:
        if hasattr(other, "id"):
            other = other.id
        return self.id == other

    def __ne__(self, other: "SnowflakeObject") -> bool:
        return not self.__eq__(other)

    def __hash__(self) -> int:
        return self.id << 32

    def __int__(self) -> int:
        return self.id

    @property
    def created_at(self) -> "models.Timestamp":
        """
        Returns a timestamp representing the date-time this discord object was created.

        :Returns:

        """
        return self.id.created_at



from typing import TYPE_CHECKING, Any, List, Mapping, cast

import discord_typings

from interactions.client.utils.serializer import dict_filter_none
from interactions.models.internal.protocols import CanRequest

from ..route import PAYLOAD_TYPE, Route

__all__ = ("GuildRequests",)


if TYPE_CHECKING:
    from interactions.models.discord.enums import AuditLogEventType
    from interactions.models.discord.snowflake import Snowflake_Type


class GuildRequests(CanRequest):
    async def get_guilds(
        self,
        limit: int = 200,
        before: "Snowflake_Type | None" = None,
        after: "Snowflake_Type | None" = None,
    ) -> list[discord_typings.GuildData]:
        """
        Get a list of partial guild objects the current user is a member of req. `guilds` scope.

        Args:
            limit: max number of guilds to return (1-200)
            before: get guilds before this guild ID
            after: get guilds after this guild ID

        Returns:
            List of guild objects

        """
        params: PAYLOAD_TYPE = {
            "limit": limit,
            "before": int(before) if before else None,
            "after": int(after) if after else None,
        }
        params = dict_filter_none(params)

        result = await self.request(Route("GET", "/users/@me/guilds", params=params))
        return cast(list[discord_typings.GuildData], result)

    async def get_guild(self, guild_id: "Snowflake_Type", with_counts: bool = True) -> discord_typings.GuildData:
        """
        Get the guild object for the given ID.

        Args:
            guild_id: the id of the guild
            with_counts: when `true`, will return approximate member and presence counts for the guild
        Returns:
            a guild object

        """
        params = {"with_counts": int(with_counts)}
        result = await self.request(Route("GET", "/guilds/{guild_id}", guild_id=guild_id), params=params)
        return cast(discord_typings.GuildData, result)

    async def get_guild_preview(self, guild_id: "Snowflake_Type") -> discord_typings.GuildPreviewData:
        """
        Get a guild's preview.

        Args:
            guild_id: the guilds ID

        Returns:
            guild preview object

        """
        result = await self.request(Route("GET", "/guilds/{guild_id}/preview", guild_id=guild_id))
        return cast(discord_typings.GuildPreviewData, result)

    async def get_channels(self, guild_id: "Snowflake_Type") -> list[discord_typings.ChannelData]:
        """
        Get a guilds channels.

        Args:
            guild_id: the id of the guild

        Returns:
            List of channels

        """
        result = await self.request(Route("GET", "/guilds/{guild_id}/channels", guild_id=guild_id))
        return cast(list[discord_typings.ChannelData], result)

    async def get_roles(self, guild_id: "Snowflake_Type") -> list[discord_typings.RoleData]:
        """
        Get a guild's roles.

        Args:
            guild_id: The ID of the guild

        Returns:
            List of roles

        """
        result = await self.request(Route("GET", "/guilds/{guild_id}/roles", guild_id=guild_id))
        return cast(list[discord_typings.RoleData], result)

    async def modify_guild(
        self, guild_id: "Snowflake_Type", reason: str | None = None, **kwargs: Mapping[str, Any]
    ) -> None:
        """
        Modify a guild's attributes.

        Args:
            guild_id: The ID of the guild we want to modify
            reason: The reason for this change
            **kwargs: The params to change

        """
        expected = (
            "name",
            "region",
            "verification_level",
            "default_message_notifications",
            "explicit_content_filter",
            "afk_channel_id",
            "afk_timeout",
            "icon",
            "owner_id",
            "splash",
            "discovery_splash",
            "banner",
            "system_channel_id",
            "system_channel_flags",
            "rules_channel_id",
            "public_updates_channel_id",
            "safety_alerts_channel_id",
            "preferred_locale",
            "features",
            "description",
        )
        payload = kwargs.copy()
        for key, value in kwargs.items():
            if key not in expected or value is None:
                del payload[key]

        # only do the request if there is something to modify
        if payload:
            await self.request(Route("PATCH", "/guilds/{guild_id}", guild_id=guild_id), payload=payload, reason=reason)

    async def delete_guild(self, guild_id: "Snowflake_Type") -> None:
        """
        Delete the guild.

        Args:
            guild_id: The ID of the guild that we want to delete

        """
        await self.request(Route("DELETE", "/guilds/{guild_id}", guild_id=guild_id))

    async def add_guild_member(
        self,
        guild_id: "Snowflake_Type",
        user_id: "Snowflake_Type",
        access_token: str,
        nick: str | None = None,
        roles: list["Snowflake_Type"] | None = None,
        mute: bool = False,
        deaf: bool = False,
    ) -> discord_typings.GuildMemberData:
        """
        Add a user to the guild. All parameters to this endpoint except for `access_token`, `guild_id` and `user_id` are optional.

        Args:
            guild_id: The ID of the guild
            user_id: The ID of the user to add
            access_token: The access token of the user
            nick: value to set users nickname to
            roles: array of role ids the member is assigned
            mute: whether the user is muted in voice channels
            deaf: whether the user is deafened in voice channels
        Returns:
            Guild Member Object

        """
        payload = {
            "access_token": access_token,
            "nick": nick,
            "roles": [int(role) for role in roles] if roles else None,
            "mute": mute,
            "deaf": deaf,
        }
        payload = dict_filter_none(payload)

        result = await self.request(
            Route("PUT", "/guilds/{guild_id}/members/{user_id}", guild_id=guild_id, user_id=user_id), payload=payload
        )
        return cast(discord_typings.GuildMemberData, result)

    async def remove_guild_member(
        self, guild_id: "Snowflake_Type", user_id: "Snowflake_Type", reason: str | None = None
    ) -> None:
        """
        Remove a member from a guild.

        Args:
            guild_id: The ID of the guild
            user_id: The ID of the user to remove
            reason: The reason for this action

        """
        await self.request(
            Route("DELETE", "/guilds/{guild_id}/members/{user_id}", guild_id=guild_id, user_id=user_id), reason=reason
        )

    async def get_guild_bans(
        self,
        guild_id: "Snowflake_Type",
        before: "Snowflake_Type | None" = None,
        after: "Snowflake_Type | None" = None,
        limit: int = 1000,
    ) -> list[discord_typings.BanData]:
        """
        Return a list of ban objects for the users banned from this guild.

        Args:
            guild_id: The ID of the guild to query
            before: snowflake to get entries before
            after: snowflake to get entries after
            limit: max number of entries to get

        Returns:
            List of ban objects

        """
        params: PAYLOAD_TYPE = {
            "limit": limit,
            "before": int(before) if before else None,
            "after": int(after) if after else None,
        }
        params = dict_filter_none(params)

        result = await self.request(Route("GET", "/guilds/{guild_id}/bans", guild_id=guild_id), params=params)
        return cast(list[discord_typings.BanData], result)

    async def get_guild_ban(self, guild_id: "Snowflake_Type", user_id: "Snowflake_Type") -> discord_typings.BanData:
        """
        Returns a ban object for the given user or a 404 not found if the ban cannot be found.

        Args:
            guild_id: The ID of the guild to query
            user_id: The ID of the user to query

        Returns:
            Ban object if exists

        Raises:
            NotFound: if no ban exists

        """
        result = await self.request(
            Route("GET", "/guilds/{guild_id}/bans/{user_id}", guild_id=guild_id, user_id=user_id)
        )
        return cast(discord_typings.BanData, result)

    async def create_guild_ban(
        self,
        guild_id: "Snowflake_Type",
        user_id: "Snowflake_Type",
        delete_message_seconds: int = 0,
        reason: str | None = None,
    ) -> None:
        """
        Create a guild ban, and optionally delete previous messages sent by the banned user.

        Args:
            guild_id: The ID of the guild to create the ban in
            user_id: The ID of the user to ban
            delete_message_seconds: number of seconds to delete messages for (0-604800)
            reason: The reason for this action

        """
        payload = {"delete_message_seconds": delete_message_seconds}
        await self.request(
            Route("PUT", "/guilds/{guild_id}/bans/{user_id}", guild_id=guild_id, user_id=user_id),
            payload=payload,
            reason=reason,
        )

    async def remove_guild_ban(
        self, guild_id: "Snowflake_Type", user_id: "Snowflake_Type", reason: str | None = None
    ) -> None:
        """
        Remove a guild ban.

        Args:
            guild_id: The ID of the guild to remove the ban in
            user_id: The ID of the user to unban
            reason: The reason for this action

        """
        await self.request(
            Route("DELETE", "/guilds/{guild_id}/bans/{user_id}", guild_id=guild_id, user_id=user_id), reason=reason
        )

    async def bulk_guild_ban(
        self,
        guild_id: "Snowflake_Type",
        user_ids: "list[Snowflake_Type]",
        delete_message_seconds: int = 0,
        reason: str | None = None,
    ) -> discord_typings.BulkBanData:
        """
        Ban a list of users from the guild.

        Args:
            guild_id: The ID of the guild to create the ban in
            user_ids: List of user ids to ban (max 200)
            delete_message_seconds: Number of seconds to delete messages for (0-604800)
            reason: The reason for this action

        Returns:
            Bulk ban object

        """
        payload = {"delete_message_days": delete_message_seconds, "user_ids": user_ids}
        result = await self.request(
            Route("POST", "/guilds/{guild_id}/bulk-ban", guild_id=guild_id), payload=payload, reason=reason
        )
        return cast(discord_typings.BulkBanData, result)

    async def get_guild_prune_count(
        self,
        guild_id: "Snowflake_Type",
        days: int = 7,
        include_roles: list["Snowflake_Type"] | None = None,
    ) -> dict:
        """
        Returns an object with one 'pruned' key indicating the number of members that would be removed in a prune operation.

        Args:
            guild_id: The ID of the guild to query
            days: number of days to count prune for (1-30)
            include_roles: role(s) to include

        Returns:
            {"pruned": int}

        """
        params: PAYLOAD_TYPE = {
            "days": days,
            "include_roles": ", ".join(str(int(role)) for role in include_roles) if include_roles else None,
        }
        params = dict_filter_none(params)

        result = await self.request(Route("GET", "/guilds/{guild_id}/prune", guild_id=guild_id), params=params)
        return cast(dict, result)  # todo revisit, create TypedDict for pruned

    async def begin_guild_prune(
        self,
        guild_id: "Snowflake_Type",
        days: int = 7,
        include_roles: list["Snowflake_Type"] | None = None,
        compute_prune_count: bool = True,
        reason: str | None = None,
    ) -> dict:
        """
        Begin a prune operation.

        Args:
            guild_id: The ID of the guild to query
            days: number of days to count prune for (1-30)
            include_roles: role(s) to include
            compute_prune_count: whether 'pruned' is returned, discouraged for large guilds
            reason: The reason for this action

        Returns:
            {"pruned": Optional[int]}

        """
        payload: PAYLOAD_TYPE = {
            "days": days,
            "compute_prune_count": compute_prune_count,
            "include_roles": ", ".join(str(int(role)) for role in include_roles) if include_roles else None,
        }
        payload = dict_filter_none(payload)

        result = await self.request(
            Route("POST", "/guilds/{guild_id}/prune", guild_id=guild_id), payload=payload, reason=reason
        )
        return cast(dict, result)  # todo revisit, create TypedDict for pruned

    async def get_guild_invites(self, guild_id: "Snowflake_Type") -> list[discord_typings.InviteData]:
        """
        Returns a list of invite objects (with invite metadata) for the guild.

        Args:
            guild_id: The ID of the guild to query

        Returns:
            List of invite objects

        """
        result = await self.request(Route("GET", "/guilds/{guild_id}/invites", guild_id=guild_id))
        return cast(list[discord_typings.InviteData], result)

    async def create_guild_role(
        self, guild_id: "Snowflake_Type", payload: dict, reason: str | None = None
    ) -> discord_typings.RoleData:
        """
        Create a new role for the guild.

        Args:
            guild_id: The ID of the guild
            payload: A dict representing the role to add
            reason: The reason for this action

        Returns:
            Role object

        """
        result = await self.request(
            Route("POST", "/guilds/{guild_id}/roles", guild_id=guild_id), payload=payload, reason=reason
        )
        return cast(discord_typings.RoleData, result)

    async def modify_guild_role_positions(
        self,
        guild_id: "Snowflake_Type",
        position_changes: List[dict["Snowflake_Type", int]],
        reason: str | None = None,
    ) -> list[discord_typings.RoleData]:
        """
        Modify the position of a role in the guild.

        Args:
            guild_id: The ID of the guild
            position_changes: A list of dicts representing the roles to move and their new positions
            reason: The reason for this action

        Returns:
            List of guild roles

        """
        payload: PAYLOAD_TYPE = [
            {"id": int(role["id"]), "position": int(role["position"])} for role in position_changes
        ]
        result = await self.request(
            Route("PATCH", "/guilds/{guild_id}/roles", guild_id=guild_id), payload=payload, reason=reason
        )
        return cast(list[discord_typings.RoleData], result)

    async def modify_guild_role(
        self,
        guild_id: "Snowflake_Type",
        role_id: "Snowflake_Type",
        payload: dict,
        reason: str | None = None,
    ) -> discord_typings.RoleData:
        """
        Modify an existing role for the guild.

        Args:
            guild_id: The ID of the guild
            role_id: The ID of the role to move
            payload: A dict representing the role to add
            reason: The reason for this action

        Returns:
            Role object

        """
        result = await self.request(
            Route("PATCH", "/guilds/{guild_id}/roles/{role_id}", guild_id=guild_id, role_id=role_id),
            payload=payload,
            reason=reason,
        )
        return cast(discord_typings.RoleData, result)

    async def delete_guild_role(
        self, guild_id: "Snowflake_Type", role_id: "Snowflake_Type", reason: str | None = None
    ) -> None:
        """
        Delete a guild role.

        Args:
            role_id: The ID of the role to delete
            reason: The reason for this action
            guild_id: The ID of the guild

        """
        await self.request(
            Route("DELETE", "/guilds/{guild_id}/roles/{role_id}", guild_id=guild_id, role_id=role_id), reason=reason
        )

    async def get_audit_log(
        self,
        guild_id: "Snowflake_Type",
        user_id: "Snowflake_Type | None" = None,
        action_type: "AuditLogEventType | None" = None,
        before: "Snowflake_Type | None" = None,
        after: "Snowflake_Type | None" = None,
        limit: int = 100,
    ) -> discord_typings.AuditLogData:
        """
        Get the audit log for a guild.

        Args:
            guild_id: The ID of the guild to query
            user_id: filter by user ID
            action_type: filter by action type
            before: snowflake to get entries before
            after: snowflake to get entries after
            limit: max number of entries to get

        Returns:
            audit log object for the guild

        """
        params: PAYLOAD_TYPE = {
            "limit": limit,
            "before": int(before) if before else None,
            "after": int(after) if after else None,
            "user_id": int(user_id) if user_id else None,
            "action_type": int(action_type) if action_type else None,
        }
        params = dict_filter_none(params)

        result = await self.request(Route("GET", "/guilds/{guild_id}/audit-logs", guild_id=guild_id), params=params)
        return cast(discord_typings.AuditLogData, result)

    async def get_guild_voice_regions(self, guild_id: "Snowflake_Type") -> list[discord_typings.VoiceRegionData]:
        """
        Returns a list of voice region objects for the guild. Unlike the similar /voice route, this returns VIP servers when the guild is VIP- enabled.

        Args:
            guild_id: The ID of the guild to query

        Returns:
            List of voice region objects

        """
        result = await self.request(Route("GET", "/guilds/{guild_id}/regions", guild_id=guild_id))
        return cast(list[discord_typings.VoiceRegionData], result)

    async def get_guild_integrations(self, guild_id: "Snowflake_Type") -> list[discord_typings.IntegrationData]:
        """
        Returns a list of integration objects for the guild.

        Args:
            guild_id: The ID of the guild to query

        Returns:
            list of integration objects

        """
        result = await self.request(Route("GET", "/guilds/{guild_id}/integrations", guild_id=guild_id))
        return cast(list[discord_typings.IntegrationData], result)

    async def delete_guild_integration(
        self,
        guild_id: "Snowflake_Type",
        integration_id: "Snowflake_Type",
        reason: str | None = None,
    ) -> None:
        """
        Delete an integration from the guild.

        Args:
            guild_id: The ID of the guild
            integration_id: The ID of the integration to remove
            reason: The reason for this action

        """
        await self.request(
            Route(
                "DELETE",
                "/guilds/{guild_id}/integrations/{integration_id}",
                guild_id=guild_id,
                integration_id=integration_id,
            ),
            reason=reason,
        )

    async def get_guild_widget_settings(self, guild_id: "Snowflake_Type") -> discord_typings.GuildWidgetSettingsData:
        """
        Get guild widget settings.

        Args:
            guild_id: The ID of the guild to query

        Returns:
            guild widget object

        """
        result = await self.request(Route("GET", "/guilds/{guild_id}/widget", guild_id=guild_id))
        return cast(discord_typings.GuildWidgetSettingsData, result)

    async def get_guild_widget(self, guild_id: "Snowflake_Type") -> discord_typings.GuildWidgetData:
        """
        Returns the widget for the guild.

        Args:
            guild_id: The ID of the guild to query

        Returns:
            Guild widget

        """
        result = await self.request(Route("GET", "/guilds/{guild_id}/widget.json", guild_id=guild_id))
        return cast(discord_typings.GuildWidgetData, result)

    async def get_guild_widget_image(self, guild_id: "Snowflake_Type", style: str | None = None) -> str:
        """
        Get a url representing a png image widget for the guild.

        For styles see: https://discord.com/developers/docs/resources/guild#get-guild-widget-image

        Args:
            guild_id: The guild to query
            style: The style of widget required.

        Returns:
            A url pointing to this image

        """
        route = Route(
            "GET", "/guilds/{guild_id}/widget.png{style}", guild_id=guild_id, style="?style={style}" if style else ""
        )
        return route.url

    async def get_guild_welcome_screen(self, guild_id: "Snowflake_Type") -> discord_typings.WelcomeScreenData:
        """
        Get the welcome screen for this guild.

        Args:
            guild_id: The ID of the guild to query
        Returns:
            Welcome screen object

        """
        result = await self.request(Route("GET", "/guilds/{guild_id}/welcome-screen", guild_id=guild_id))
        return cast(discord_typings.WelcomeScreenData, result)

    async def get_guild_vanity_url(self, guild_id: "Snowflake_Type") -> dict:
        """
        Get a partial invite object for the guilds vanity invite url.

        Args:
            guild_id: The ID of the guild to query

        Returns:
            Returns a partial invite object. Code is None if a vanity url for the guild is not set.

        """
        result = await self.request(Route("GET", "/guilds/{guild_id}/vanity-url", guild_id=guild_id))
        return cast(dict, result)  # todo create typing?

    async def get_guild_channels(
        self, guild_id: "Snowflake_Type"
    ) -> list[discord_typings.ChannelData]:  # todo narrow down channel types
        """
        Gets a list of guild channel objects.

        Args:
            guild_id: The ID of the guild

        Returns:
            A list of channels in this guild. Does not include threads.

        """
        result = await self.request(Route("GET", "/guilds/{guild_id}/channels", guild_id=guild_id))
        return cast(list[discord_typings.ChannelData], result)

    async def modify_guild_widget(
        self,
        guild_id: "Snowflake_Type",
        enabled: bool | None = None,
        channel_id: "Snowflake_Type | None" = None,
    ) -> discord_typings.GuildWidgetData:
        """
        Modify a guild widget.

        Args:
            guild_id: The ID of the guild to modify.
            enabled: Should the guild widget be enabled
            channel_id: The widget's channel ID

        Returns:
            Updated guild widget.

        """
        payload: PAYLOAD_TYPE = {
            "enabled": enabled,
            "channel_id": int(channel_id) if channel_id else None,
        }
        payload = dict_filter_none(payload)

        result = await self.request(Route("PATCH", "/guilds/{guild_id}/widget", guild_id=guild_id), payload=payload)
        return cast(discord_typings.GuildWidgetData, result)

    async def modify_guild_welcome_screen(
        self,
        guild_id: "Snowflake_Type",
        enabled: bool,
        welcome_channels: list["Snowflake_Type"],
        description: str,
    ) -> discord_typings.WelcomeScreenData:
        """
        Modify the guild's welcome screen.

        Args:
            guild_id: The ID of the guild.
            enabled: Whether the welcome screen is enabled
            welcome_channels: Channels linked in the welcome screen and their display options
            description: The server description to show in the welcome screen

        Returns:
            Updated welcome screen object

        """
        payload: PAYLOAD_TYPE = {
            "enabled": enabled,
            "welcome_channels": [int(channel) for channel in welcome_channels],
            "description": description,
        }
        result = await self.request(
            Route("PATCH", "/guilds/{guild_id}/welcome-screen", guild_id=guild_id), payload=payload
        )
        return cast(discord_typings.WelcomeScreenData, result)

    async def modify_current_user_voice_state(
        self,
        guild_id: "Snowflake_Type",
        channel_id: "Snowflake_Type",
        suppress: bool | None = None,
        request_to_speak_timestamp: str | None = None,
    ) -> None:
        """
        Update the current user voice state.

        Args:
            guild_id: The ID of the guild to update.
            channel_id: The id of the channel the user is currently in
            suppress: Toggle the user's suppress state.
            request_to_speak_timestamp: Sets the user's request to speak

        """
        payload: PAYLOAD_TYPE = {
            "suppress": suppress,
            "request_to_speak_timestamp": request_to_speak_timestamp,
            "channel_id": int(channel_id) if channel_id else None,
        }
        payload = dict_filter_none(payload)

        await self.request(Route("PATCH", "/guilds/{guild_id}/voice-states/@me", guild_id=guild_id), payload=payload)

    async def modify_user_voice_state(
        self,
        guild_id: "Snowflake_Type",
        user_id: "Snowflake_Type",
        channel_id: "Snowflake_Type",
        suppress: bool | None = None,
    ) -> None:
        """
        Modify the voice state of a user.

        Args:
            guild_id: The ID of the guild.
            user_id: The ID of the user to modify.
            channel_id: The ID of the channel the user is currently in.
            suppress: Toggles the user's suppress state.

        """
        payload: PAYLOAD_TYPE = {
            "suppress": suppress,
            "channel_id": int(channel_id) if channel_id else None,
        }
        payload = dict_filter_none(payload)

        await self.request(
            Route("PATCH", "/guilds/{guild_id}/voice-states/{user_id}", guild_id=guild_id, user_id=user_id),
            payload=payload,
        )

    async def create_guild(
        self,
        name: str,
        icon: str | None = None,
        verification_level: int | None = None,
        default_message_notifications: int | None = None,
        explicit_content_filter: int | None = None,
        roles: list[dict] | None = None,
        channels: list[dict] | None = None,
        afk_channel_id: "Snowflake_Type | None" = None,
        afk_timeout: int | None = None,
        system_channel_id: "Snowflake_Type | None" = None,
        system_channel_flags: int | None = None,
    ) -> discord_typings.GuildData:
        payload = {
            "name": name,
            "icon": icon,
            "verification_level": verification_level,
            "default_message_notifications": default_message_notifications,
            "explicit_content_filter": explicit_content_filter,
            "roles": roles,
            "channels": channels,
            "afk_channel_id": int(afk_channel_id) if afk_channel_id else None,
            "afk_timeout": afk_timeout,
            "system_channel_id": int(system_channel_id) if system_channel_id else None,
            "system_channel_flags": system_channel_flags,
        }
        payload = dict_filter_none(payload)

        result = await self.request(Route("POST", "/guilds"), payload=payload)
        return cast(discord_typings.GuildData, result)

    async def create_guild_from_guild_template(
        self, template_code: str, name: str, icon: str
    ) -> discord_typings.GuildData:
        """
        Creates a new guild based on a template.

        !!! note
            This endpoint can only be used by bots in less than 10 guilds.

        Args:
            template_code: The code of the template to use.
            name: The name of the guild (2-100 characters)
            icon: Data URI scheme

        Returns:
            The newly created guild object

        """
        payload = {"name": name, "icon": icon}

        result = await self.request(
            Route("POST", "/guilds/templates/{template_code}", template_code=template_code), payload=payload
        )
        return cast(discord_typings.GuildData, result)

    async def get_guild_templates(self, guild_id: "Snowflake_Type") -> list[discord_typings.GuildTemplateData]:
        """
        Returns an array of guild templates.

        Args:
            guild_id: The ID of the guild to query.

        Returns:
            An array of guild templates

        """
        result = await self.request(Route("GET", "/guilds/{guild_id}/templates", guild_id=guild_id))
        return cast(list[discord_typings.GuildTemplateData], result)

    async def create_guild_template(
        self, guild_id: "Snowflake_Type", name: str, description: str | None = None
    ) -> discord_typings.GuildTemplateData:
        """
        Create a guild template for the guild.

        Args:
            guild_id: The ID of the guild to create a template for.
            name: The name of the template
            description: The description of the template

        Returns:
            The created guild template

        """
        payload = {"name": name, "description": description}
        payload = dict_filter_none(payload)

        result = await self.request(Route("POST", "/guilds/{guild_id}/templates", guild_id=guild_id), payload=payload)
        return cast(discord_typings.GuildTemplateData, result)

    async def sync_guild_template(
        self, guild_id: "Snowflake_Type", template_code: str
    ) -> discord_typings.GuildTemplateData:
        """
        Sync the template to the guild's current state.

        Args:
            guild_id: The ID of the guild
            template_code: The code for the template to sync

        Returns:
            The updated guild template

        """
        result = await self.request(
            Route("PUT", "/guilds/{guild_id}/templates/{template_code}", guild_id=guild_id, template_code=template_code)
        )
        return cast(discord_typings.GuildTemplateData, result)

    async def modify_guild_template(
        self,
        guild_id: "Snowflake_Type",
        template_code: str,
        name: str | None = None,
        description: str | None = None,
    ) -> discord_typings.GuildTemplateData:
        """
        Modifies the template's metadata.

        Args:
            guild_id: The ID of the guild
            template_code: The template code
            name: The name of the template
            description: The description of the template

        Returns:
            The updated guild template

        """
        payload: PAYLOAD_TYPE = {"name": name, "description": description}
        payload = dict_filter_none(payload)

        result = await self.request(
            Route(
                "PATCH", "/guilds/{guild_id}/templates/{template_code}", guild_id=guild_id, template_code=template_code
            ),
            payload=payload,
        )
        return cast(discord_typings.GuildTemplateData, result)

    async def delete_guild_template(
        self, guild_id: "Snowflake_Type", template_code: str
    ) -> discord_typings.GuildTemplateData:
        """
        Delete the guild template.

        Args:
            guild_id: The ID of the guild
            template_code: The ID of the template

        Returns:
            The deleted template object

        """
        # why on earth does this return the deleted template object?
        result = await self.request(
            Route(
                "DELETE", "/guilds/{guild_id}/templates/{template_code}", guild_id=guild_id, template_code=template_code
            )
        )
        return cast(discord_typings.GuildTemplateData, result)

    async def get_auto_moderation_rules(
        self, guild_id: "Snowflake_Type"
    ) -> list[discord_typings.AutoModerationRuleData]:
        """
        Get this guilds auto moderation rules.

        Args:
            guild_id: The ID of the guild to get

        Returns:
            A list of auto moderation rules

        """
        result = await self.request(Route("GET", "/guilds/{guild_id}/auto-moderation/rules", guild_id=guild_id))
        return cast(list[dict], result)

    async def get_auto_moderation_rule(
        self, guild_id: "Snowflake_Type", rule_id: "Snowflake_Type"
    ) -> discord_typings.AutoModerationRuleData:
        """
        Get a specific auto moderation rule.

        Args:
            guild_id: The ID of the guild
            rule_id: The ID of the rule to get

        Returns:
            The auto moderation rule

        """
        result = await self.request(
            Route("GET", "/guilds/{guild_id}/auto-moderation/rules/{rule_id}", guild_id=guild_id, rule_id=rule_id)
        )
        return cast(dict, result)

    async def create_auto_moderation_rule(
        self, guild_id: "Snowflake_Type", payload: discord_typings.AutoModerationRuleData
    ) -> discord_typings.AutoModerationRuleData:
        """
        Create an auto moderation rule.

        Args:
            guild_id: The ID of the guild to create this rule within
            payload: A dict representing the auto moderation rule

        Returns:
            The created auto moderation rule

        """
        result = await self.request(
            Route("POST", "/guilds/{guild_id}/auto-moderation/rules", guild_id=guild_id), payload=payload
        )
        return cast(dict, result)

    async def modify_auto_moderation_rule(
        self,
        guild_id: "Snowflake_Type",
        rule_id: "Snowflake_Type",
        name: str | None = None,
        trigger_type: dict | None = None,
        trigger_metadata: dict | None = None,
        actions: list[dict] | None = None,
        exempt_channels: list["Snowflake_Type"] | None = None,
        exempt_roles: list["Snowflake_Type"] | None = None,
        event_type: dict | None = None,
        enabled: bool | None = None,
        reason: str | None = None,
    ) -> dict:
        """
        Modify an existing auto moderation rule.

        Args:
            guild_id: The ID of the guild the rule belongs to
            rule_id: The ID of the rule to modify
            name: The name of the rule
            trigger_type: The type trigger for this rule
            trigger_metadata: Metadata for the trigger
            actions: A list of actions to take upon triggering
            exempt_roles: Roles that ignore this rule
            exempt_channels: Channels that ignore this role
            enabled: Is this rule enabled?
            event_type: The type of event that triggers this rule
            reason: The reason for this change

        Returns:
            The updated rule object

        """
        payload = {
            "name": name,
            "trigger_type": trigger_type,
            "trigger_metadata": trigger_metadata,
            "actions": actions,
            "exempt_roles": [int(role) for role in exempt_roles] if exempt_roles else None,
            "exempt_channels": [int(channel) for channel in exempt_channels] if exempt_channels else None,
            "event_type": event_type,
            "enabled": enabled,
        }
        payload = dict_filter_none(payload)

        result = await self.request(
            Route("PATCH", "/guilds/{guild_id}/auto-moderation/rules/{rule_id}", guild_id=guild_id, rule_id=rule_id),
            payload=payload,
            reason=reason,
        )
        return cast(dict, result)

    async def delete_auto_moderation_rule(
        self, guild_id: "Snowflake_Type", rule_id: "Snowflake_Type", reason: str | None = None
    ) -> dict:
        """
        Delete an auto moderation rule.

        Args:
            guild_id: The ID of the guild to delete this rule from
            rule_id: The ID of the role to delete
            reason: The reason for deleting this rule

        """
        result = await self.request(
            Route("DELETE", "/guilds/{guild_id}/auto-moderation/rules/{rule_id}", guild_id=guild_id, rule_id=rule_id),
            reason=reason,
        )
        return cast(dict, result)

    async def get_guild_onboarding(self, guild_id: "Snowflake_Type") -> discord_typings.GuildOnboardingData:
        """
        Get the guild's onboarding settings.

        Args:
            guild_id: The ID of the guild

        Returns:
            The guild's onboarding object

        """
        result = await self.request(Route("GET", "/guilds/{guild_id}/onboarding", guild_id=guild_id))
        return cast(discord_typings.GuildOnboardingData, result)

    async def modify_guild_onboarding(
        self, guild_id: "Snowflake_Type", payload: dict, reason: str | None = None
    ) -> discord_typings.GuildOnboardingData:
        """
        Modify the guild's onboarding settings.

        Args:
            guild_id: The ID of the guild
            payload: A dict representing the modified Onboarding
            reason: The reason for this action

        Returns:
            The updated onboarding object

        """
        result = await self.request(
            Route("PUT", "/guilds/{guild_id}/onboarding", guild_id=guild_id),
            payload=payload,
            reason=reason,
        )
        return cast(discord_typings.GuildOnboardingData, result)



from typing import TYPE_CHECKING

from interactions.models.discord.entitlement import Entitlement
import interactions.api.events as events
from ._template import EventMixinTemplate, Processor

if TYPE_CHECKING:
    from interactions.api.events import RawGatewayEvent


class EntitlementEvents(EventMixinTemplate):
    @Processor.define()
    async def _on_raw_entitlement_create(self, event: "RawGatewayEvent") -> None:
        self.dispatch(events.EntitlementCreate(Entitlement.from_dict(event.data, self)))

    @Processor.define()
    async def _on_raw_entitlement_update(self, event: "RawGatewayEvent") -> None:
        self.dispatch(events.EntitlementUpdate(Entitlement.from_dict(event.data, self)))

    @Processor.define()
    async def _on_raw_entitlement_delete(self, event: "RawGatewayEvent") -> None:
        self.dispatch(events.EntitlementDelete(Entitlement.from_dict(event.data, self)))



import copy
from typing import TYPE_CHECKING

import interactions.api.events as events
from interactions.client.const import MISSING
from ._template import EventMixinTemplate, Processor

if TYPE_CHECKING:
    from interactions.api.events import RawGatewayEvent

__all__ = ("RoleEvents",)


class RoleEvents(EventMixinTemplate):
    @Processor.define()
    async def _on_raw_guild_role_create(self, event: "RawGatewayEvent") -> None:
        g_id = int(event.data.get("guild_id"))
        r_id = int(event.data["role"]["id"])

        if guild := self.cache.get_guild(g_id):
            guild._role_ids.add(r_id)

        role = self.cache.place_role_data(g_id, [event.data.get("role")])[r_id]
        self.dispatch(events.RoleCreate(g_id, role))

    @Processor.define()
    async def _on_raw_guild_role_update(self, event: "RawGatewayEvent") -> None:
        g_id = int(event.data.get("guild_id"))
        r_data = event.data.get("role")
        before = copy.copy(self.cache.get_role(r_data["id"]) or MISSING)

        after = self.cache.place_role_data(g_id, [r_data])
        after = after[int(event.data["role"]["id"])]

        self.dispatch(events.RoleUpdate(g_id, before, after))

    @Processor.define()
    async def _on_raw_guild_role_delete(self, event: "RawGatewayEvent") -> None:
        g_id = int(event.data.get("guild_id"))
        r_id = int(event.data.get("role_id"))

        role = self.cache.get_role(r_id)

        self.cache.delete_role(r_id)

        if guild := self.cache.get_guild(g_id):
            role_members = (member for member in guild.members if member.has_role(r_id))
            for member in role_members:
                member._role_ids.remove(r_id)

        self.dispatch(events.RoleDelete(g_id, r_id, role))



import asyncio
import base64
import re
from collections import namedtuple
from dataclasses import dataclass
from typing import (
    TYPE_CHECKING,
    Any,
    AsyncGenerator,
    Dict,
    List,
    Mapping,
    Optional,
    Sequence,
    Union,
)

import attrs

import interactions.models as models
from interactions.client.const import GUILD_WELCOME_MESSAGES, MISSING, Absent
from interactions.client.errors import NotFound, ThreadOutsideOfGuild
from interactions.client.mixins.serialization import DictSerializationMixin
from interactions.client.utils.attr_converters import optional as optional_c
from interactions.client.utils.attr_converters import timestamp_converter
from interactions.client.utils.serializer import dict_filter_none
from interactions.client.utils.text_utils import mentions
from interactions.models.discord.channel import BaseChannel, GuildChannel
from interactions.models.discord.embed import process_embeds
from interactions.models.discord.emoji import process_emoji_req_format
from interactions.models.discord.file import UPLOADABLE_TYPE
from interactions.models.discord.poll import Poll
from interactions.models.misc.iterator import AsyncIterator

from .base import DiscordObject
from .enums import (
    AutoArchiveDuration,
    ChannelType,
    InteractionType,
    MentionType,
    MessageActivityType,
    MessageFlags,
    MessageType,
    IntegrationType,
)
from .snowflake import (
    Snowflake,
    Snowflake_Type,
    to_optional_snowflake,
    to_snowflake,
    to_snowflake_list,
)

if TYPE_CHECKING:
    from interactions import InteractionContext
    from interactions.client import Client

__all__ = (
    "Attachment",
    "ChannelMention",
    "MessageActivity",
    "MessageReference",
    "MessageInteraction",
    "MessageInteractionMetadata",
    "AllowedMentions",
    "BaseMessage",
    "Message",
    "MessageType",
    "process_allowed_mentions",
    "process_message_reference",
    "process_message_payload",
)

channel_mention = re.compile(r"<#(?P<id>[0-9]{17,})>")


class PollAnswerVotersIterator(AsyncIterator):
    def __init__(
        self, message: "Message", answer_id: int, limit: int = 25, after: Snowflake_Type | None = None
    ) -> None:
        self.message: "Message" = message
        self.answer_id = answer_id
        self.after: Snowflake_Type | None = after
        self._more: bool = True
        super().__init__(limit)

    async def fetch(self) -> list["models.User"]:
        if not self.last:
            self.last = namedtuple("temp", "id")
            self.last.id = self.after

        rcv = await self.message._client.http.get_answer_voters(
            self.message._channel_id,
            self.message.id,
            self.answer_id,
            limit=self.get_limit,
            after=to_snowflake(self.last.id) if self.last.id else None,
        )
        if not rcv:
            raise asyncio.QueueEmpty

        users = [self.message._client.cache.place_user_data(user_data) for user_data in rcv["users"]]
        return users


@attrs.define(eq=False, order=False, hash=False, kw_only=True)
class Attachment(DiscordObject):
    filename: str = attrs.field(
        repr=False,
    )
    """name of file attached"""
    description: Optional[str] = attrs.field(repr=False, default=None)
    """description for the file"""
    content_type: Optional[str] = attrs.field(repr=False, default=None)
    """the attachment's media type"""
    size: int = attrs.field(
        repr=False,
    )
    """size of file in bytes"""
    url: str = attrs.field(
        repr=False,
    )
    """source url of file"""
    proxy_url: str = attrs.field(
        repr=False,
    )
    """a proxied url of file"""
    height: Optional[int] = attrs.field(repr=False, default=None)
    """height of file (if image)"""
    width: Optional[int] = attrs.field(repr=False, default=None)
    """width of file (if image)"""
    ephemeral: bool = attrs.field(repr=False, default=False)
    """whether this attachment is ephemeral"""
    duration_secs: Optional[int] = attrs.field(repr=False, default=None)
    """the duration of the audio file (currently for voice messages)"""
    waveform: bytearray = attrs.field(repr=False, default=None)
    """base64 encoded bytearray representing a sampled waveform (currently for voice messages)"""

    @property
    def resolution(self) -> tuple[Optional[int], Optional[int]]:
        """Returns the image resolution of the attachment file"""
        return self.height, self.width

    @classmethod
    def _process_dict(cls, data: Dict[str, Any], _) -> Dict[str, Any]:
        if waveform := data.pop("waveform", None):
            data["waveform"] = bytearray(base64.b64decode(waveform))
        return data


@attrs.define(eq=False, order=False, hash=False, kw_only=True)
class ChannelMention(DiscordObject):
    guild_id: "Snowflake_Type | None" = attrs.field(
        repr=False,
    )
    """id of the guild containing the channel"""
    type: ChannelType = attrs.field(repr=False, converter=ChannelType)
    """the type of channel"""
    name: str = attrs.field(
        repr=False,
    )
    """the name of the channel"""


@dataclass
class MessageActivity:
    type: MessageActivityType
    """type of message activity"""
    party_id: str = None
    """party_id from a Rich Presence event"""


@attrs.define(eq=False, order=False, hash=False, kw_only=True)
class MessageReference(DictSerializationMixin):
    """
    Reference to an originating message.

    Can be used for replies.

    """

    message_id: int = attrs.field(repr=False, default=None, converter=optional_c(to_snowflake))
    """id of the originating message."""
    channel_id: Optional[int] = attrs.field(repr=False, default=None, converter=optional_c(to_snowflake))
    """id of the originating message's channel."""
    guild_id: Optional[int] = attrs.field(repr=False, default=None, converter=optional_c(to_snowflake))
    """id of the originating message's guild."""
    fail_if_not_exists: bool = attrs.field(repr=False, default=True)
    """When sending a message, whether to error if the referenced message doesn't exist instead of sending as a normal (non-reply) message, default true."""

    @classmethod
    def for_message(cls, message: "Message", fail_if_not_exists: bool = True) -> "MessageReference":
        """
        Creates a reference to a message.

        Parameters
            message: The target message to reference.
            fail_if_not_exists: Whether to error if the referenced message doesn't exist instead of sending as a normal (non-reply) message

        Returns:
            A MessageReference object.

        """
        return cls(
            message_id=message.id,
            channel_id=message._channel_id,
            guild_id=message._guild_id,
            fail_if_not_exists=fail_if_not_exists,
        )


@attrs.define(eq=False, order=False, hash=False, kw_only=True)
class MessageInteraction(DiscordObject):
    type: InteractionType = attrs.field(repr=False, converter=InteractionType)
    """the type of interaction"""
    name: str = attrs.field(
        repr=False,
    )
    """the name of the application command"""

    _user_id: "Snowflake_Type" = attrs.field(
        repr=False,
    )
    """ID of the user who triggered the interaction"""

    @classmethod
    def _process_dict(cls, data: Dict[str, Any], client: "Client") -> Dict[str, Any]:
        user_data = data["user"]
        data["user_id"] = client.cache.place_user_data(user_data).id
        return data

    @property
    def user(self) -> "models.User":
        """Get the user associated with this interaction."""
        return self.client.get_user(self._user_id)


@attrs.define(eq=False, order=False, hash=False, kw_only=True)
class MessageInteractionMetadata(DiscordObject):
    type: InteractionType = attrs.field(repr=False, converter=InteractionType)
    """The type of interaction"""
    authorizing_integration_owners: dict[IntegrationType, Snowflake] = attrs.field(repr=False, factory=dict)
    """IDs for installation context(s) related to an interaction."""
    original_response_message_id: "Optional[Snowflake_Type]" = attrs.field(
        repr=False, default=None, converter=to_optional_snowflake
    )
    """ID of the original response message, present only on follow-up messages"""
    interacted_message_id: "Optional[Snowflake_Type]" = attrs.field(
        repr=False, default=None, converter=to_optional_snowflake
    )
    """ID of the message that contained interactive component, present only on messages created from component interactions"""
    triggering_interaction_metadata: "Optional[MessageInteractionMetadata]" = attrs.field(repr=False, default=None)
    """Metadata for the interaction that was used to open the modal, present only on modal submit interactions"""

    _user_id: "Snowflake_Type" = attrs.field(
        repr=False,
    )
    """ID of the user who triggered the interaction"""

    @classmethod
    def _process_dict(cls, data: Dict[str, Any], client: "Client") -> Dict[str, Any]:
        if "authorizing_integration_owners" in data:
            data["authorizing_integration_owners"] = {
                IntegrationType(int(integration_type)): Snowflake(owner_id)
                for integration_type, owner_id in data["authorizing_integration_owners"].items()
            }
        if "triggering_interaction_metadata" in data:
            data["triggering_interaction_metadata"] = cls.from_dict(data["triggering_interaction_metadata"], client)

        user_data = data["user"]
        data["user_id"] = client.cache.place_user_data(user_data).id

        return data

    @property
    def user(self) -> "models.User":
        """Get the user associated with this interaction."""
        return self.client.get_user(self.user_id)


@attrs.define(eq=False, order=False, hash=False, kw_only=False)
class AllowedMentions(DictSerializationMixin):
    """
    The allowed mention field allows for more granular control over mentions without various hacks to the message content.

    This will always validate against message content to avoid phantom
    pings, and check against user/bot permissions.

    """

    parse: Optional[List[str]] = attrs.field(repr=False, factory=list)
    """An array of allowed mention types to parse from the content."""
    roles: Optional[List["Snowflake_Type"]] = attrs.field(repr=False, factory=list, converter=to_snowflake_list)
    """Array of role_ids to mention. (Max size of 100)"""
    users: Optional[List["Snowflake_Type"]] = attrs.field(repr=False, factory=list, converter=to_snowflake_list)
    """Array of user_ids to mention. (Max size of 100)"""
    replied_user = attrs.field(repr=False, default=False)
    """For replies, whether to mention the author of the message being replied to. (default false)"""

    def add_parse(self, *mention_types: Union["MentionType", str]) -> None:
        """
        Add a mention type to the list of allowed mentions to parse.

        Args:
            *mention_types: The types of mentions to add

        """
        for mention_type in mention_types:
            if not isinstance(mention_type, MentionType) and mention_type not in MentionType.__members__.values():
                raise ValueError(f"Invalid mention type: {mention_type}")
            self.parse.append(mention_type)

    def add_roles(self, *roles: Union["models.Role", "Snowflake_Type"]) -> None:
        """
        Add roles that are allowed to be mentioned.

        Args:
            *roles: The roles to add

        """
        for role in roles:
            self.roles.append(to_snowflake(role))

    def add_users(self, *users: Union["models.Member", "models.BaseUser", "Snowflake_Type"]) -> None:
        """
        Add users that are allowed to be mentioned.

        Args:
            *users: The users to add

        """
        for user in users:
            self.users.append(to_snowflake(user))

    @classmethod
    def all(cls) -> "AllowedMentions":
        """
        Allows every user and role to be mentioned.

        Returns:
            An AllowedMentions object

        """
        return cls(parse=list(MentionType.__members__.values()), replied_user=True)

    @classmethod
    def none(cls) -> "AllowedMentions":
        """
        Disallows any user or role to be mentioned.

        Returns:
            An AllowedMentions object

        """
        return cls()


@attrs.define(eq=False, order=False, hash=False, kw_only=True)
class BaseMessage(DiscordObject):
    _channel_id: "Snowflake_Type" = attrs.field(repr=False, default=MISSING, converter=to_optional_snowflake)
    _thread_channel_id: Optional["Snowflake_Type"] = attrs.field(
        repr=False, default=None, converter=to_optional_snowflake
    )
    _guild_id: Optional["Snowflake_Type"] = attrs.field(repr=False, default=None, converter=to_optional_snowflake)
    _author_id: "Snowflake_Type" = attrs.field(repr=False, default=MISSING, converter=to_optional_snowflake)

    @property
    def guild(self) -> "models.Guild":
        """The guild the message was sent in"""
        return self._client.cache.get_guild(self._guild_id)

    @property
    def channel(self) -> "models.TYPE_MESSAGEABLE_CHANNEL":
        """The channel the message was sent in"""
        channel = self._client.cache.get_channel(self._channel_id)

        if not self._guild_id and not channel:
            # allow dm operations without fetching a dm channel from API
            channel = BaseChannel.from_dict_factory({"id": self._channel_id, "type": ChannelType.DM}, self._client)
            if self.author:
                channel.recipients = [self.author]
        return channel

    @property
    def thread(self) -> "models.TYPE_THREAD_CHANNEL":
        """The thread that was started from this message, includes thread member object"""
        return self._client.cache.get_channel(self._thread_channel_id)

    @property
    def author(self) -> Union["models.Member", "models.User"]:
        """The author of this message. Only a valid user in the case where the message is generated by a user or bot user."""
        if self._author_id:
            member = None
            if self._guild_id:
                member = self._client.cache.get_member(self._guild_id, self._author_id)
            return member or self._client.cache.get_user(self._author_id)
        return MISSING


@attrs.define(eq=False, order=False, hash=False, kw_only=True)
class Message(BaseMessage):
    content: str = attrs.field(repr=False, default=MISSING)
    """Contents of the message"""
    timestamp: "models.Timestamp" = attrs.field(repr=False, default=MISSING, converter=optional_c(timestamp_converter))
    """When this message was sent"""
    edited_timestamp: Optional["models.Timestamp"] = attrs.field(
        repr=False, default=None, converter=optional_c(timestamp_converter)
    )
    """When this message was edited (or `None` if never)"""
    tts: bool = attrs.field(repr=False, default=False)
    """Whether this was a TTS message"""
    mention_everyone: bool = attrs.field(repr=False, default=False)
    """Whether this message mentions everyone"""
    mention_channels: List[ChannelMention] = attrs.field(repr=False, factory=list)
    """Channels specifically mentioned in this message"""
    attachments: List[Attachment] = attrs.field(repr=False, factory=list)
    """Any attached files"""
    embeds: List["models.Embed"] = attrs.field(repr=False, factory=list)
    """Any embedded content"""
    reactions: List["models.Reaction"] = attrs.field(repr=False, factory=list)
    """Reactions to the message"""
    nonce: Optional[Union[int, str]] = attrs.field(repr=False, default=None)
    """Used for validating a message was sent"""
    pinned: bool = attrs.field(repr=False, default=False)
    """Whether this message is pinned"""
    webhook_id: Optional["Snowflake_Type"] = attrs.field(repr=False, default=None, converter=to_optional_snowflake)
    """If the message is generated by a webhook, this is the webhook's id"""
    type: MessageType = attrs.field(repr=False, default=MISSING, converter=optional_c(MessageType))
    """Type of message"""
    activity: Optional[MessageActivity] = attrs.field(repr=False, default=None, converter=optional_c(MessageActivity))
    """Activity sent with Rich Presence-related chat embeds"""
    application: Optional["models.Application"] = attrs.field(repr=False, default=None)  # TODO: partial application
    """Application sent with Rich Presence-related chat embeds"""
    application_id: Optional["Snowflake_Type"] = attrs.field(repr=False, default=None, converter=to_optional_snowflake)
    """If the message is an Interaction or application-owned webhook, this is the id of the application"""
    message_reference: Optional[MessageReference] = attrs.field(
        repr=False, default=None, converter=optional_c(MessageReference.from_dict)
    )
    """Data showing the source of a crosspost, channel follow add, pin, or reply message"""
    flags: MessageFlags = attrs.field(repr=False, default=MessageFlags.NONE, converter=MessageFlags)
    """Message flags combined as a bitfield"""
    poll: Optional[Poll] = attrs.field(repr=False, default=None, converter=optional_c(Poll.from_dict))
    """A poll."""
    interaction_metadata: Optional[MessageInteractionMetadata] = attrs.field(repr=False, default=None)
    """Sent if the message is a response to an Interaction"""
    interaction: Optional["MessageInteraction"] = attrs.field(repr=False, default=None)
    """(Deprecated in favor of interaction_metadata) Sent if the message is a response to an Interaction"""
    components: Optional[List["models.ActionRow"]] = attrs.field(repr=False, default=None)
    """Sent if the message contains components like buttons, action rows, or other interactive components"""
    sticker_items: Optional[List["models.StickerItem"]] = attrs.field(repr=False, default=None)
    """Sent if the message contains stickers"""
    _mention_ids: List["Snowflake_Type"] = attrs.field(repr=False, factory=list)
    _mention_roles: List["Snowflake_Type"] = attrs.field(repr=False, factory=list)
    _referenced_message_id: Optional["Snowflake_Type"] = attrs.field(repr=False, default=None)

    @property
    async def mention_users(self) -> AsyncGenerator[Union["models.Member", "models.User"], None]:
        """A generator of users mentioned in this message"""
        for u_id in self._mention_ids:
            if self._guild_id:
                yield await self._client.cache.fetch_member(self._guild_id, u_id)
            else:
                yield await self._client.cache.fetch_user(u_id)

    @property
    async def mention_roles(self) -> AsyncGenerator["models.Role", None]:
        """A generator of roles mentioned in this message"""
        for r_id in self._mention_roles:
            yield await self._client.cache.fetch_role(self._guild_id, r_id)

    @property
    def thread(self) -> "models.TYPE_THREAD_CHANNEL":
        """The thread that was started from this message, if any"""
        return self._client.cache.get_channel(self.id)

    @property
    def editable(self) -> bool:
        """Whether this message can be edited by the current user"""
        if self.author.id == self._client.user.id:
            return MessageFlags.VOICE_MESSAGE not in self.flags
        return False

    async def fetch_referenced_message(self, *, force: bool = False) -> Optional["Message"]:
        """
        Fetch the message this message is referencing, if any.

        Args:
            force: Whether to force a fetch from the API

        Returns:
            The referenced message, if found

        """
        if self._referenced_message_id is None:
            return None

        try:
            return await self._client.cache.fetch_message(self._channel_id, self._referenced_message_id, force=force)
        except NotFound:
            return None

    def get_referenced_message(self) -> Optional["Message"]:
        """
        Get the message this message is referencing, if any.

        Returns:
            The referenced message, if found

        """
        if self._referenced_message_id is None:
            return None
        return self._client.cache.get_message(self._channel_id, self._referenced_message_id)

    def contains_mention(
        self,
        query: "str | re.Pattern[str] | models.BaseUser | models.BaseChannel | models.Role",
        *,
        tag_as_mention: bool = False,
    ) -> bool:
        """
        Check whether the message contains the query or not.

        Args:
            query: The query to search for
            tag_as_mention: Should `BaseUser.tag` be checked *(only if query is an instance of BaseUser)*

        Returns:
            A boolean indicating whether the query could be found or not

        """
        return mentions(text=self.content or self.system_content, query=query, tag_as_mention=tag_as_mention)

    @classmethod
    def _process_dict(cls, data: dict, client: "Client") -> dict:  # noqa: C901
        if author_data := data.pop("author", None):
            if "guild_id" in data and "member" in data:
                author_data["member"] = data.pop("member")
                data["author_id"] = client.cache.place_member_data(data["guild_id"], author_data).id
            else:
                data["author_id"] = client.cache.place_user_data(author_data).id

        if mentions_data := data.pop("mentions", None):
            mention_ids = []
            for user_data in mentions_data:
                if "guild_id" in data and "member" in user_data:
                    mention_ids.append(client.cache.place_member_data(data["guild_id"], user_data).id)
                else:
                    mention_ids.append(client.cache.place_user_data(user_data).id)
            data["mention_ids"] = mention_ids

        found_ids = []
        mention_channels = []
        if "mention_channels" in data:
            for channel_data in data["mention_channels"]:
                mention_channels.append(ChannelMention.from_dict(channel_data, client))
                found_ids.append(channel_data["id"])
        if "content" in data:
            for channel_id in channel_mention.findall(data["content"]):
                if channel_id not in found_ids and (channel := client.get_channel(channel_id)):
                    channel_data = {
                        "id": channel.id,
                        "guild_id": channel._guild_id if isinstance(channel, GuildChannel) else None,
                        "type": channel.type,
                        "name": channel.name,
                    }
                    mention_channels.append(ChannelMention.from_dict(channel_data, client))
        if mention_channels:
            data["mention_channels"] = mention_channels

        if "attachments" in data:
            data["attachments"] = Attachment.from_list(data.get("attachments"), client)

        if "embeds" in data:
            data["embeds"] = models.Embed.from_list(data.get("embeds"))

        if "reactions" in data:
            reactions = [
                models.Reaction.from_dict(
                    reaction_data | {"message_id": data["id"], "channel_id": data["channel_id"]},
                    client,
                )
                for reaction_data in data["reactions"]
            ]
            data["reactions"] = reactions

        # TODO: Convert to application object

        if ref_message_data := data.pop("referenced_message", None):
            if not ref_message_data.get("guild_id"):
                ref_message_data["guild_id"] = data.get("guild_id")
            _m = client.cache.place_message_data(ref_message_data)
            data["referenced_message_id"] = _m.id
        elif msg_reference := data.get("message_reference"):
            data["referenced_message_id"] = msg_reference.get("message_id")

        if "interaction_metadata" in data:
            data["interaction_metadata"] = MessageInteractionMetadata.from_dict(data["interaction_metadata"], client)

        if "interaction" in data:
            data["interaction"] = MessageInteraction.from_dict(data["interaction"], client)

        if thread_data := data.pop("thread", None):
            data["thread_channel_id"] = client.cache.place_channel_data(thread_data).id

        if "components" in data:
            components = [
                models.BaseComponent.from_dict_factory(component_data) for component_data in data["components"]
            ]
            data["components"] = components

        if "sticker_items" in data:
            data["sticker_items"] = models.StickerItem.from_list(data["sticker_items"], client)

        return data

    @property
    def system_content(self) -> Optional[str]:
        """Content for system messages. (boosts, welcomes, etc)"""
        match self.type:
            case MessageType.USER_PREMIUM_GUILD_SUBSCRIPTION:
                return f"{self.author.mention} just boosted the server!"
            case MessageType.USER_PREMIUM_GUILD_SUBSCRIPTION_TIER_1:
                return f"{self.author.mention} just boosted the server! {self.guild.name} has achieved **Level 1!**"
            case MessageType.USER_PREMIUM_GUILD_SUBSCRIPTION_TIER_2:
                return f"{self.author.mention} just boosted the server! {self.guild.name} has achieved **Level 2!**"
            case MessageType.USER_PREMIUM_GUILD_SUBSCRIPTION_TIER_3:
                return f"{self.author.mention} just boosted the server! {self.guild.name} has achieved **Level 3!**"
            case MessageType.GUILD_MEMBER_JOIN:
                return GUILD_WELCOME_MESSAGES[
                    int(self.timestamp.timestamp() * 1000)
                    % len(GUILD_WELCOME_MESSAGES)
                    # This is how Discord calculates the welcome message.
                ].format(self.author.mention)
            case MessageType.THREAD_CREATED:
                return f"{self.author.mention} started a thread: {self.thread.mention}. See all **threads**."
            case MessageType.CHANNEL_FOLLOW_ADD:
                return f"{self.author.mention} has added **{self.content}** to this channel. Its most important updates will show up here."
            case MessageType.RECIPIENT_ADD:
                return f"{self.author.mention} added <@{self._mention_ids[0]}> to the thread."
            case MessageType.RECIPIENT_REMOVE:
                return f"{self.author.mention} removed <@{self._mention_ids[0]}> from the thread."
            case MessageType.CHANNEL_NAME_CHANGE:
                return f"{self.author.mention} changed the channel name: **{self.content}**."
            case MessageType.CHANNEL_PINNED_MESSAGE:
                return f"{self.author.mention} pinned a message. See all pinned messages"
            case MessageType.GUILD_DISCOVERY_DISQUALIFIED:
                return "This server has been removed from Server Discovery because it no longer passes all the requirements. Check Server Settings for more details."
            case MessageType.GUILD_DISCOVERY_REQUALIFIED:
                return "This server is eligible for Server Discovery again and has been automatically relisted!"
            case MessageType.GUILD_DISCOVERY_GRACE_PERIOD_INITIAL_WARNING:
                return "This server has failed Discovery activity requirements for 1 week. If this server fails for 4 weeks in a row, it will be automatically removed from Discovery."
            case MessageType.GUILD_DISCOVERY_GRACE_PERIOD_FINAL_WARNING:
                return "This server has failed Discovery activity requirements for 3 weeks in a row. If this server fails for 1 more week, it will be removed from Discovery."
            case MessageType.GUILD_INVITE_REMINDER:
                return "**Invite your friends**\nThe best way to setup a server is with your buddies!"
            case MessageType.THREAD_STARTER_MESSAGE:
                if referenced_message := self.get_referenced_message():
                    return referenced_message.content
                return "Sorry, we couldn't load the first message in this thread"
            case MessageType.AUTO_MODERATION_ACTION:
                keyword_matched_content = self.embeds[0].fields[4].value  # The words that triggered the action
                message_content = self.embeds[0].description.replace(
                    keyword_matched_content, f"**{keyword_matched_content}**"
                )
                rule = self.embeds[0].fields[0].value  # What rule was triggered
                channel = self.embeds[0].fields[1].value  # Channel that the action took place in
                return f'AutoMod has blocked a message in <#{channel}>. "{message_content}" from {self.author.mention}. Rule: {rule}.'
            case _:
                return None

    @property
    def jump_url(self) -> str:
        """A url that allows the client to *jump* to this message."""
        return f"https://discord.com/channels/{self._guild_id or '@me'}/{self._channel_id}/{self.id}"

    @property
    def proto_url(self) -> str:
        """A URL like `jump_url` that uses protocols."""
        return f"discord://-/channels/{self._guild_id or '@me'}/{self._channel_id}/{self.id}"

    def answer_voters(
        self, answer_id: int, limit: int = 0, before: Snowflake_Type | None = None
    ) -> PollAnswerVotersIterator:
        """
        An async iterator for getting the voters for an answer in the poll this message has.

        Args:
            answer_id: The answer to get voters for
            after: Get messages after this user ID
            limit: The max number of users to return (default 25, max 100)

        """
        return PollAnswerVotersIterator(self, answer_id, limit, before)

    async def edit(
        self,
        *,
        content: Optional[str] = None,
        embeds: Optional[Union[Sequence[Union["models.Embed", dict]], Union["models.Embed", dict]]] = None,
        embed: Optional[Union["models.Embed", dict]] = None,
        components: Optional[
            Union[
                Sequence[Sequence[Union["models.BaseComponent", dict]]],
                Sequence[Union["models.BaseComponent", dict]],
                "models.BaseComponent",
                dict,
            ]
        ] = None,
        allowed_mentions: Optional[Union[AllowedMentions, dict]] = None,
        attachments: Optional[Optional[Sequence[Union[Attachment, dict]]]] = None,
        files: Optional[Union[UPLOADABLE_TYPE, Sequence[UPLOADABLE_TYPE]]] = None,
        file: Optional[UPLOADABLE_TYPE] = None,
        tts: bool = False,
        flags: Optional[Union[int, MessageFlags]] = None,
        context: "InteractionContext | None" = None,
    ) -> "Message":
        """
        Edits the message.

        Args:
            content: Message text content.
            embeds: Embedded rich content (up to 6000 characters).
            embed: Embedded rich content (up to 6000 characters).
            components: The components to include with the message.
            allowed_mentions: Allowed mentions for the message.
            attachments: The attachments to keep, only used when editing message.
            files: Files to send, the path, bytes or File() instance, defaults to None. You may have up to 10 files.
            file: Files to send, the path, bytes or File() instance, defaults to None. You may have up to 10 files.
            tts: Should this message use Text To Speech.
            flags: Message flags to apply.
            context: The interaction context to use for the edit

        Returns:
            New message object with edits applied

        """
        if context:
            return await context.edit(
                self,
                content=content,
                embeds=embeds,
                embed=embed,
                components=components,
                allowed_mentions=allowed_mentions,
                attachments=attachments,
                files=files,
                file=file,
                tts=tts,
            )
        message_payload = process_message_payload(
            content=content,
            embeds=embed if embeds is None else embeds,
            components=components,
            allowed_mentions=allowed_mentions,
            attachments=attachments,
            tts=tts,
            flags=flags,
        )
        if file:
            files = [file, *files] if files else [file]
        message_data = await self._client.http.edit_message(message_payload, self._channel_id, self.id, files=files)
        if message_data:
            return self._client.cache.place_message_data(message_data)

    async def delete(self, delay: int = 0, *, context: "InteractionContext | None" = None) -> None:
        """
        Delete message.

        Args:
            delay: Seconds to wait before deleting message.
            context: An optional interaction context to delete ephemeral messages.

        """

        async def _delete() -> None:
            if delay:
                await asyncio.sleep(delay)

            if MessageFlags.EPHEMERAL in self.flags:
                if not context:
                    raise ValueError("Cannot delete ephemeral message without interaction context parameter")
                await context.delete(self.id)
            else:
                await self._client.http.delete_message(self._channel_id, self.id)

        if delay:
            return asyncio.create_task(_delete())

        return await _delete()

    async def reply(
        self,
        content: Optional[str] = None,
        embeds: Optional[Union[List[Union["models.Embed", dict]], Union["models.Embed", dict]]] = None,
        embed: Optional[Union["models.Embed", dict]] = None,
        **kwargs: Mapping[str, Any],
    ) -> "Message":
        """
        Reply to this message, takes all the same attributes as `send`.

        Args:
            content: Message text content.
            embeds: Embedded rich content (up to 6000 characters).
            embed: Embedded rich content (up to 6000 characters).
            **kwargs: Additional options to pass to `send`.

        Returns:
            New message object.

        """
        return await self.channel.send(content=content, reply_to=self, embeds=embeds or embed, **kwargs)

    async def create_thread(
        self,
        name: str,
        auto_archive_duration: Union[AutoArchiveDuration, int] = AutoArchiveDuration.ONE_DAY,
        reason: Optional[str] = None,
    ) -> "models.TYPE_THREAD_CHANNEL":
        """
        Create a thread from this message.

        Args:
            name: The name of this thread
            auto_archive_duration: duration in minutes to automatically archive the thread after recent activity, can be set to: 60, 1440, 4320, 10080
            reason: The optional reason for creating this thread

        Returns:
            The created thread object

        Raises:
            ThreadOutsideOfGuild: if this is invoked on a message outside of a guild

        """
        if self.channel.type not in (ChannelType.GUILD_TEXT, ChannelType.GUILD_NEWS):
            raise ThreadOutsideOfGuild

        thread_data = await self._client.http.create_thread(
            channel_id=self._channel_id,
            name=name,
            auto_archive_duration=auto_archive_duration,
            message_id=self.id,
            reason=reason,
        )
        return self._client.cache.place_channel_data(thread_data)

    async def suppress_embeds(self) -> "Message":
        """
        Suppress embeds for this message.

        !!! note
            Requires the `Permissions.MANAGE_MESSAGES` permission.

        """
        message_data = await self._client.http.edit_message(
            {"flags": MessageFlags.SUPPRESS_EMBEDS}, self._channel_id, self.id
        )
        if message_data:
            return self._client.cache.place_message_data(message_data)

    async def fetch_reaction(
        self,
        emoji: Union["models.PartialEmoji", dict, str],
        limit: Absent[int] = MISSING,
        after: Absent["Snowflake_Type"] = MISSING,
    ) -> List["models.User"]:
        """
        Fetches reactions of a specific emoji from this message.

        Args:
            emoji: The emoji to get
            limit: Max number of users to return (1-100)
            after: Get users after this user ID

        Returns:
            list of users who have reacted with that emoji

        """
        reaction_data = await self._client.http.get_reactions(
            self._channel_id, self.id, process_emoji_req_format(emoji), limit, to_optional_snowflake(after)
        )
        return [self._client.cache.place_user_data(user_data) for user_data in reaction_data]

    async def add_reaction(self, emoji: Union["models.PartialEmoji", dict, str]) -> None:
        """
        Add a reaction to this message.

        Args:
            emoji: the emoji to react with

        """
        emoji = process_emoji_req_format(emoji)
        await self._client.http.create_reaction(self._channel_id, self.id, emoji)

    async def remove_reaction(
        self,
        emoji: Union["models.PartialEmoji", dict, str],
        member: Optional[Union["models.Member", "models.User", "Snowflake_Type"]] = MISSING,
    ) -> None:
        """
        Remove a specific reaction that a user reacted with.

        Args:
            emoji: Emoji to remove
            member: Member to remove reaction of. Default's to ClientUser.

        """
        emoji_str = process_emoji_req_format(emoji)
        if not member:
            member = self._client.user
        user_id = to_snowflake(member)
        if user_id == self._client.user.id:
            await self._client.http.remove_self_reaction(self._channel_id, self.id, emoji_str)
        else:
            await self._client.http.remove_user_reaction(self._channel_id, self.id, emoji_str, user_id)

    async def clear_reactions(self, emoji: Union["models.PartialEmoji", dict, str]) -> None:
        """
        Clear a specific reaction from message.

        Args:
            emoji: The emoji to clear

        """
        emoji = models.process_emoji_req_format(emoji)
        await self._client.http.clear_reaction(self._channel_id, self.id, emoji)

    async def clear_all_reactions(self) -> None:
        """Clear all emojis from a message."""
        await self._client.http.clear_reactions(self._channel_id, self.id)

    async def pin(self) -> None:
        """Pin message."""
        await self._client.http.pin_message(self._channel_id, self.id)
        self.pinned = True

    async def unpin(self) -> None:
        """Unpin message."""
        await self._client.http.unpin_message(self._channel_id, self.id)
        self.pinned = False

    async def publish(self) -> None:
        """
        Publish this message.

        (Discord api calls it "crosspost")

        """
        await self._client.http.crosspost_message(self._channel_id, self.id)

    async def end_poll(self) -> "Message":
        """Ends the poll contained in this message."""
        message_data = await self._client.http.end_poll(self._channel_id, self.id)
        if message_data:
            return self._client.cache.place_message_data(message_data)


def process_allowed_mentions(allowed_mentions: Optional[Union[AllowedMentions, dict]]) -> Optional[dict]:
    """
    Process allowed mentions into a dictionary.

    Args:
        allowed_mentions: Allowed mentions object or dictionary

    Returns:
        Dictionary of allowed mentions

    Raises:
        ValueError: Invalid allowed mentions

    """
    if not allowed_mentions:
        return allowed_mentions

    if isinstance(allowed_mentions, dict):
        return allowed_mentions

    if isinstance(allowed_mentions, AllowedMentions):
        return allowed_mentions.to_dict()

    raise ValueError(f"Invalid allowed mentions: {allowed_mentions}")


def process_message_reference(
    message_reference: Optional[Union[MessageReference, Message, dict, "Snowflake_Type"]]
) -> Optional[dict]:
    """
    Process mention references into a dictionary.

    Args:
        message_reference: Message reference object

    Returns:
        Message reference dictionary

    Raises:
        ValueError: Invalid message reference

    """
    if not message_reference:
        return message_reference

    if isinstance(message_reference, dict):
        return message_reference

    if isinstance(message_reference, (str, int)):
        message_reference = MessageReference(message_id=message_reference)

    if isinstance(message_reference, Message):
        message_reference = MessageReference.for_message(message_reference)

    if isinstance(message_reference, MessageReference):
        return message_reference.to_dict()

    raise ValueError(f"Invalid message reference: {message_reference}")


def process_message_payload(
    content: Optional[str] = None,
    embeds: Optional[Union[List[Union["models.Embed", dict]], Union["models.Embed", dict]]] = None,
    components: Optional[
        Union[
            List[List[Union["models.BaseComponent", dict]]],
            List[Union["models.BaseComponent", dict]],
            "models.BaseComponent",
            dict,
        ]
    ] = None,
    stickers: Optional[
        Union[List[Union["models.Sticker", "Snowflake_Type"]], "models.Sticker", "Snowflake_Type"]
    ] = None,
    allowed_mentions: Optional[Union[AllowedMentions, dict]] = None,
    reply_to: Optional[Union[MessageReference, Message, dict, "Snowflake_Type"]] = None,
    attachments: Optional[List[Union[Attachment, dict]]] = None,
    tts: bool = False,
    flags: Optional[Union[int, MessageFlags]] = None,
    nonce: Optional[str | int] = None,
    enforce_nonce: bool = False,
    poll: Optional[Poll | dict] = None,
    **kwargs,
) -> dict:
    """
    Format message content for it to be ready to send discord.

    Args:
        content: Message text content.
        embeds: Embedded rich content (up to 6000 characters).
        components: The components to include with the message.
        stickers: IDs of up to 3 stickers in the server to send in the message.
        allowed_mentions: Allowed mentions for the message.
        reply_to: Message to reference, must be from the same channel.
        attachments: The attachments to keep, only used when editing message.
        tts: Should this message use Text To Speech.
        flags: Message flags to apply.
        nonce: Used to verify a message was sent.
        enforce_nonce: If enabled and nonce is present, it will be checked for uniqueness in the past few minutes. \
            If another message was created by the same author with the same nonce, that message will be returned \
            and no new message will be created.
        poll: A poll.

    Returns:
        Dictionary

    """
    embeds = process_embeds(embeds)
    if isinstance(embeds, list):
        embeds = embeds if all(e is not None for e in embeds) else None

    components = models.process_components(components)
    if stickers:
        stickers = [to_snowflake(sticker) for sticker in stickers]
    allowed_mentions = process_allowed_mentions(allowed_mentions)
    message_reference = process_message_reference(reply_to)
    if attachments:
        attachments = [attachment.to_dict() for attachment in attachments]

    if isinstance(poll, Poll):
        poll = poll.to_dict()

    return dict_filter_none(
        {
            "content": content,
            "embeds": embeds,
            "components": components,
            "sticker_ids": stickers,
            "allowed_mentions": allowed_mentions,
            "message_reference": message_reference,
            "attachments": attachments,
            "tts": tts,
            "flags": flags,
            "nonce": nonce,
            "enforce_nonce": enforce_nonce,
            "poll": poll,
            **kwargs,
        }
    )



from typing import TYPE_CHECKING

from interactions.models.discord.auto_mod import AutoModerationAction, AutoModRule

from ... import events
from ._template import EventMixinTemplate, Processor

if TYPE_CHECKING:
    from interactions.api.events import RawGatewayEvent

__all__ = ("AutoModEvents",)


class AutoModEvents(EventMixinTemplate):
    @Processor.define()
    async def _raw_auto_moderation_action_execution(self, event: "RawGatewayEvent") -> None:
        action = AutoModerationAction.from_dict(event.data.copy(), self)
        channel = self.get_channel(event.data.get("channel_id"))
        guild = self.get_guild(event.data["guild_id"])
        self.dispatch(events.AutoModExec(action, channel, guild))

    @Processor.define()
    async def raw_auto_moderation_rule_create(self, event: "RawGatewayEvent") -> None:
        rule = AutoModRule.from_dict(event.data, self)
        guild = self.get_guild(event.data["guild_id"])
        self.dispatch(events.AutoModCreated(guild, rule))

    @Processor.define()
    async def raw_auto_moderation_rule_update(self, event: "RawGatewayEvent") -> None:
        rule = AutoModRule.from_dict(event.data, self)
        guild = self.get_guild(event.data["guild_id"])
        self.dispatch(events.AutoModUpdated(guild, rule))

    @Processor.define()
    async def raw_auto_moderation_rule_delete(self, event: "RawGatewayEvent") -> None:
        rule = AutoModRule.from_dict(event.data, self)
        guild = self.get_guild(event.data["guild_id"])
        self.dispatch(events.AutoModDeleted(guild, rule))



from typing import TYPE_CHECKING, Any, Iterable, Optional, Union

import interactions.models as models
import interactions.models.discord
from interactions.models.discord.enums import MessageFlags

if TYPE_CHECKING:
    from interactions.client import Client
    from interactions.models.discord.file import UPLOADABLE_TYPE
    from interactions.models.discord.components import BaseComponent
    from interactions.models.discord.embed import Embed
    from interactions.models.discord.message import AllowedMentions, Message, MessageReference
    from interactions.models.discord.poll import Poll
    from interactions.models.discord.sticker import Sticker
    from interactions.models.discord.snowflake import Snowflake_Type


__all__ = ("SendMixin",)


class SendMixin:
    client: "Client"

    async def _send_http_request(self, message_payload: dict, files: Iterable["UPLOADABLE_TYPE"] | None = None) -> dict:
        raise NotImplementedError

    async def send(
        self,
        content: Optional[str] = None,
        *,
        embeds: Optional[Union[Iterable[Union["Embed", dict]], Union["Embed", dict]]] = None,
        embed: Optional[Union["Embed", dict]] = None,
        components: Optional[
            Union[
                Iterable[Iterable[Union["BaseComponent", dict]]],
                Iterable[Union["BaseComponent", dict]],
                "BaseComponent",
                dict,
            ]
        ] = None,
        stickers: Optional[Union[Iterable[Union["Sticker", "Snowflake_Type"]], "Sticker", "Snowflake_Type"]] = None,
        allowed_mentions: Optional[Union["AllowedMentions", dict]] = None,
        reply_to: Optional[Union["MessageReference", "Message", dict, "Snowflake_Type"]] = None,
        files: Optional[Union["UPLOADABLE_TYPE", Iterable["UPLOADABLE_TYPE"]]] = None,
        file: Optional["UPLOADABLE_TYPE"] = None,
        tts: bool = False,
        suppress_embeds: bool = False,
        silent: bool = False,
        flags: Optional[Union[int, "MessageFlags"]] = None,
        delete_after: Optional[float] = None,
        nonce: Optional[str | int] = None,
        enforce_nonce: bool = False,
        poll: "Optional[Poll | dict]" = None,
        **kwargs: Any,
    ) -> "Message":
        """
        Send a message.

        Args:
            content: Message text content.
            embeds: Embedded rich content (up to 6000 characters).
            embed: Embedded rich content (up to 6000 characters).
            components: The components to include with the message.
            stickers: IDs of up to 3 stickers in the server to send in the message.
            allowed_mentions: Allowed mentions for the message.
            reply_to: Message to reference, must be from the same channel.
            files: Files to send, the path, bytes or File() instance, defaults to None. You may have up to 10 files.
            file: Files to send, the path, bytes or File() instance, defaults to None. You may have up to 10 files.
            tts: Should this message use Text To Speech.
            suppress_embeds: Should embeds be suppressed on this send
            silent: Should this message be sent without triggering a notification.
            flags: Message flags to apply.
            delete_after: Delete message after this many seconds.
            nonce: Used to verify a message was sent. Can be up to 25 characters.
            enforce_nonce: If enabled and nonce is present, it will be checked for uniqueness in the past few minutes. \
                If another message was created by the same author with the same nonce, that message will be returned \
                and no new message will be created.
            poll: A poll.

        Returns:
            New message object that was sent.

        """
        if not flags:
            flags = MessageFlags(0)
        if suppress_embeds:
            if isinstance(flags, int):
                flags = MessageFlags(flags)
            flags = flags | MessageFlags.SUPPRESS_EMBEDS
        if silent:
            if isinstance(flags, int):
                flags = MessageFlags(flags)
            flags = flags | MessageFlags.SILENT

        if (
            files
            and (
                isinstance(files, Iterable)
                and any(isinstance(file, interactions.models.discord.message.Attachment) for file in files)
            )
            or isinstance(files, interactions.models.discord.message.Attachment)
        ):
            raise ValueError(
                "Attachments are not files. Attachments only contain metadata about the file, not the file itself - to send an attachment, you need to download it first. Check Attachment.url"
            )

        if enforce_nonce and not nonce:
            raise ValueError("You must provide a nonce to use enforce_nonce.")

        message_payload = models.discord.message.process_message_payload(
            content=content,
            embeds=embeds or embed,
            components=components,
            stickers=stickers,
            allowed_mentions=allowed_mentions,
            reply_to=reply_to,
            tts=tts,
            flags=flags,
            nonce=nonce,
            enforce_nonce=enforce_nonce,
            poll=poll,
            **kwargs,
        )

        message_data = await self._send_http_request(message_payload, files=files or file)
        if message_data:
            message = self.client.cache.place_message_data(message_data)
            if delete_after:
                if kwargs.get("pass_self_into_delete"):  # hack to pass in interaction/hybrid context
                    await message.delete(delay=delete_after, context=self)
                else:
                    await message.delete(delay=delete_after)
            return message



from datetime import datetime
from typing import Any, Dict, List, Optional, Union

import attrs
from attrs.validators import instance_of
from attrs.validators import optional as v_optional

from interactions.client.const import (
    EMBED_MAX_NAME_LENGTH,
    EMBED_MAX_FIELDS,
    EMBED_MAX_DESC_LENGTH,
    EMBED_TOTAL_MAX,
    EMBED_FIELD_VALUE_LENGTH,
)
from interactions.client.mixins.serialization import DictSerializationMixin
from interactions.client.utils.attr_converters import optional as c_optional, list_converter
from interactions.client.utils.attr_converters import timestamp_converter
from interactions.client.utils.serializer import no_export_meta, export_converter
from interactions.models.discord.color import Color, process_color
from interactions.models.discord.enums import EmbedType
from interactions.models.discord.timestamp import Timestamp

__all__ = (
    "EmbedField",
    "EmbedAuthor",
    "EmbedAttachment",
    "EmbedAuthor",
    "EmbedFooter",
    "EmbedProvider",
    "Embed",
    "process_embeds",
)


@attrs.define(eq=False, order=False, hash=False, kw_only=False)
class EmbedField(DictSerializationMixin):
    """
    Representation of an embed field.

    Attributes:
        name: Field name
        value: Field value
        inline: If the field should be inline

    """

    name: str = attrs.field(
        repr=False,
    )
    value: str = attrs.field(
        repr=False,
    )
    inline: bool = attrs.field(repr=False, default=False)

    @name.validator
    def _name_validation(self, attribute: str, value: Any) -> None:
        if len(value) > EMBED_MAX_NAME_LENGTH:
            raise ValueError(f"Field name cannot exceed {EMBED_MAX_NAME_LENGTH} characters")

    @value.validator
    def _value_validation(self, attribute: str, value: Any) -> None:
        if len(value) > EMBED_FIELD_VALUE_LENGTH:
            raise ValueError(f"Field value cannot exceed {EMBED_FIELD_VALUE_LENGTH} characters")

    def __len__(self) -> int:
        return len(self.name) + len(self.value)


@attrs.define(eq=False, order=False, hash=False, kw_only=False)
class EmbedAuthor(DictSerializationMixin):
    """
    Representation of an embed author.

    Attributes:
        name: Name to show on embed
        url: Url to go to when name is clicked
        icon_url: Icon to show next to name
        proxy_icon_url: Proxy icon url

    """

    name: Optional[str] = attrs.field(repr=False, default=None)
    url: Optional[str] = attrs.field(repr=False, default=None)
    icon_url: Optional[str] = attrs.field(repr=False, default=None)
    proxy_icon_url: Optional[str] = attrs.field(repr=False, default=None, metadata=no_export_meta)

    @name.validator
    def _name_validation(self, attribute: str, value: Any) -> None:
        if len(value) > EMBED_MAX_NAME_LENGTH:
            raise ValueError(f"Field name cannot exceed {EMBED_MAX_NAME_LENGTH} characters")

    def __len__(self) -> int:
        return len(self.name)


@attrs.define(eq=False, order=False, hash=False, kw_only=False)
class EmbedAttachment(DictSerializationMixin):  # thumbnail or image or video
    """
    Representation of an attachment.

    Attributes:
        url: Attachment url
        proxy_url: Proxy url
        height: Attachment height
        width: Attachment width

    """

    url: Optional[str] = attrs.field(repr=False, default=None)
    proxy_url: Optional[str] = attrs.field(repr=False, default=None, metadata=no_export_meta)
    height: Optional[int] = attrs.field(repr=False, default=None, metadata=no_export_meta)
    width: Optional[int] = attrs.field(repr=False, default=None, metadata=no_export_meta)

    @classmethod
    def _process_dict(cls, data: Dict[str, Any]) -> Dict[str, Any]:
        return {"url": data} if isinstance(data, str) else data

    @property
    def size(self) -> tuple[Optional[int], Optional[int]]:
        return self.height, self.width


@attrs.define(eq=False, order=False, hash=False, kw_only=False)
class EmbedFooter(DictSerializationMixin):
    """
    Representation of an Embed Footer.

    Attributes:
        text: Footer text
        icon_url: Footer icon url
        proxy_icon_url: Proxy icon url

    """

    text: str = attrs.field(
        repr=False,
    )
    icon_url: Optional[str] = attrs.field(repr=False, default=None)
    proxy_icon_url: Optional[str] = attrs.field(repr=False, default=None, metadata=no_export_meta)

    @classmethod
    def converter(cls, ingest: Union[dict, str, "EmbedFooter"]) -> "EmbedFooter":
        """
        A converter to handle users passing raw strings or dictionaries as footers to the Embed object.

        Args:
            ingest: The data to convert

        Returns:
            An EmbedFooter object

        """
        return cls(text=ingest) if isinstance(ingest, str) else cls.from_dict(ingest)

    def __len__(self) -> int:
        return len(self.text)


@attrs.define(eq=False, order=False, hash=False, kw_only=False)
class EmbedProvider(DictSerializationMixin):
    """
    Represents an embed's provider.

    !!! note
        Only used by system embeds, not bots

    Attributes:
        name: Provider name
        url: Provider url

    """

    name: Optional[str] = attrs.field(repr=False, default=None)
    url: Optional[str] = attrs.field(repr=False, default=None)


@attrs.define(eq=False, order=False, hash=False, kw_only=False)
class Embed(DictSerializationMixin):
    """Represents a discord embed object."""

    title: Optional[str] = attrs.field(default=None, repr=True)
    """The title of the embed"""
    description: Optional[str] = attrs.field(default=None, repr=True)
    """The description of the embed"""
    color: Optional[Union[Color, dict, tuple, list, str, int]] = attrs.field(
        default=None, repr=True, metadata=export_converter(process_color)
    )
    """The colour of the embed"""
    url: Optional[str] = attrs.field(default=None, validator=v_optional(instance_of(str)), repr=True)
    """The url the embed should direct to when clicked"""
    timestamp: Optional[Timestamp] = attrs.field(
        default=None,
        converter=c_optional(timestamp_converter),
        validator=v_optional(instance_of((datetime, float, int))),
        repr=True,
    )
    """Timestamp of embed content"""
    fields: List[EmbedField] = attrs.field(factory=list, converter=EmbedField.from_list, repr=True)
    """A list of [fields][interactions.models.discord.embed.EmbedField] to go in the embed"""
    author: Optional[EmbedAuthor] = attrs.field(repr=False, default=None, converter=c_optional(EmbedAuthor.from_dict))
    """The author of the embed"""
    thumbnail: Optional[EmbedAttachment] = attrs.field(
        repr=False, default=None, converter=c_optional(EmbedAttachment.from_dict)
    )
    """The thumbnail of the embed"""
    images: list[EmbedAttachment] = attrs.field(
        repr=False, factory=list, converter=list_converter(EmbedAttachment.from_dict)
    )
    """The images of the embed"""
    video: Optional[EmbedAttachment] = attrs.field(
        repr=False,
        default=None,
        converter=c_optional(EmbedAttachment.from_dict),
        metadata=no_export_meta,
    )
    """The video of the embed, only used by system embeds"""
    footer: Optional[EmbedFooter] = attrs.field(repr=False, default=None, converter=c_optional(EmbedFooter.converter))
    """The footer of the embed"""
    provider: Optional[EmbedProvider] = attrs.field(
        repr=False,
        default=None,
        converter=c_optional(EmbedProvider.from_dict),
        metadata=no_export_meta,
    )
    """The provider of the embed, only used for system embeds"""
    type: EmbedType = attrs.field(
        repr=False,
        default=EmbedType.RICH,
        converter=c_optional(EmbedType),
        metadata=no_export_meta,
    )

    @classmethod
    def _process_dict(cls, data: Dict[str, Any]) -> Dict[str, Any]:
        if image_data := data.pop("image", None):
            data["images"] = [image_data]
        return data

    @property
    def image(self) -> Optional[EmbedAttachment]:
        """
        The image of the embed.

        Raises:
            ValueError: If there are multiple images in the embed.

        """
        if len(self.images) <= 1:
            return self.images[0] if self.images else None
        raise ValueError("There are multiple images in this embed, use `images` instead")

    @image.setter
    def image(self, value: Optional[EmbedAttachment]) -> "Embed":
        """Set the image of the embed."""
        self.images = [] if value is None else [value]
        return self

    @title.validator
    def _name_validation(self, attribute: str, value: Any) -> None:
        """Validate the embed title."""
        if value is not None:
            if isinstance(value, str):
                if len(value) > EMBED_MAX_NAME_LENGTH:
                    raise ValueError(f"Title cannot exceed {EMBED_MAX_NAME_LENGTH} characters")
                return
            raise TypeError("Title must be of type String")

    @description.validator
    def _description_validation(self, attribute: str, value: Any) -> None:
        """Validate the description."""
        if value is not None:
            if isinstance(value, str):
                if len(value) > EMBED_MAX_DESC_LENGTH:
                    raise ValueError(f"Description cannot exceed {EMBED_MAX_DESC_LENGTH} characters")
                return
            raise TypeError("Description must be of type String")

    @fields.validator
    def _fields_validation(self, attribute: str, value: Any) -> None:
        """Validate the fields."""
        if isinstance(value, list) and len(value) > EMBED_MAX_FIELDS:
            raise ValueError(f"Embeds can only hold {EMBED_MAX_FIELDS} fields")

    def _check_object(self) -> None:
        self._name_validation("title", self.title)
        self._description_validation("description", self.description)
        self._fields_validation("fields", self.fields)

        if len(self) > EMBED_TOTAL_MAX:
            raise ValueError(
                "Your embed is too large, more info at https://discord.com/developers/docs/resources/channel#embed-limits"
            )

    def __len__(self) -> int:
        # yes i know there are far more optimal ways to write this
        # its written like this for readability
        total: int = 0
        if self.title:
            total += len(self.title)
        if self.description:
            total += len(self.description)
        if self.footer:
            total += len(self.footer)
        if self.author:
            total += len(self.author)
        if self.fields:
            total += sum(map(len, self.fields))
        return total

    def __bool__(self) -> bool:
        return any(
            (
                self.title,
                self.description,
                self.fields,
                self.author,
                self.thumbnail,
                self.footer,
                self.images,
                self.video,
            )
        )

    def set_author(
        self,
        name: str,
        url: Optional[str] = None,
        icon_url: Optional[str] = None,
    ) -> "Embed":
        """
        Set the author field of the embed.

        Args:
            name: The text to go in the title section
            url: A url link to the author
            icon_url: A url of an image to use as the icon

        """
        self.author = EmbedAuthor(name=name, url=url, icon_url=icon_url)
        return self

    def set_thumbnail(self, url: str) -> "Embed":
        """
        Set the thumbnail of the embed.

        Args:
            url: the url of the image to use

        """
        self.thumbnail = EmbedAttachment(url=url)
        return self

    def set_image(self, url: str) -> "Embed":
        """
        Set the image of the embed.

        Args:
            url: the url of the image to use

        """
        self.images = [EmbedAttachment(url=url)]
        return self

    def set_images(self, *images: str) -> "Embed":
        """
        Set multiple images for the embed.

        Note:
            To use multiple images, you must also set a url for this embed.

        Warning:
            This takes advantage of an undocumented feature of the API, and may be removed at any time.

        Args:
            images: the images to use

        """
        if len(self.images) + len(images) > 1 and not self.url:
            raise ValueError("To use multiple images, you must also set a url for this embed")

        self.images = [EmbedAttachment(url=url) for url in images]
        return self

    def add_image(self, image: str) -> "Embed":
        """
        Add an image to the embed.

        Note:
            To use multiple images, you must also set a url for this embed.

        Warning:
            This takes advantage of an undocumented feature of the API, and may be removed at any time.

        Args:
            image: the image to add

        """
        if len(self.images) > 0 and not self.url:
            raise ValueError("To use multiple images, you must also set a url for this embed")
        self.images.append(EmbedAttachment(url=image))
        return self

    def set_footer(self, text: str, icon_url: Optional[str] = None) -> "Embed":
        """
        Set the footer field of the embed.

        Args:
            text: The text to go in the title section
            icon_url: A url of an image to use as the icon

        """
        self.footer = EmbedFooter(text=text, icon_url=icon_url)
        return self

    def add_field(self, name: str, value: Any, inline: bool = False) -> "Embed":
        """
        Add a field to the embed.

        Args:
            name: The title of this field
            value: The value in this field
            inline: Should this field be inline with other fields?

        """
        self.fields.append(EmbedField(name, str(value), inline))
        self._fields_validation("fields", self.fields)

        return self

    def add_fields(self, *fields: EmbedField | str | dict) -> "Embed":
        """
        Add multiple fields to the embed.

        Args:
            fields: The fields to add

        """
        for _field in fields:
            if isinstance(_field, EmbedField):
                self.fields.append(_field)
                self._fields_validation("fields", self.fields)
            elif isinstance(_field, str):
                self.add_field(_field, _field)
            elif isinstance(_field, dict):
                self.add_field(**_field)
            else:
                raise TypeError(f"Expected EmbedField, str or dict, got {type(_field).__name__}")
        return self

    def to_dict(self) -> Dict[str, Any]:
        data = super().to_dict()
        if images := data.pop("images", []):
            if len(images) > 1:
                if not self.url:
                    raise ValueError("To use multiple images, you must also set a url for this embed")

                data["image"] = images[0]
                data = [data]

                data.extend({"image": image, "url": self.url} for image in images[1:])
            else:
                data["image"] = images[0]

        return data


def process_embeds(embeds: Optional[Union[List[Union[Embed, Dict]], Union[Embed, Dict]]]) -> Optional[List[dict]]:
    """
    Process the passed embeds into a format discord will understand.

    Args:
        embeds: List of dict / embeds to process

    Returns:
        formatted list for discord

    """
    if embeds is None:
        # Its just empty, so nothing to process.
        return embeds

    if isinstance(embeds, Embed):
        # Single embed, convert it to dict and wrap it into a list for discord.
        out = embeds.to_dict()
        return out if isinstance(out, list) else [out]
    if isinstance(embeds, dict):
        # We assume the dict correctly represents a single discord embed and just send it blindly
        # after wrapping it in a list for discord
        return [embeds]

    if isinstance(embeds, list):
        # A list of embeds, convert Embed to dict representation if needed.
        out = [embed.to_dict() if isinstance(embed, Embed) else embed for embed in embeds]
        if any(isinstance(embed, list) for embed in out):
            raise ValueError("You cannot send multiple embeds when using multiple images in a single embed")
        return out

    raise ValueError(f"Invalid embeds: {embeds}")



from typing import TYPE_CHECKING

import attrs

from interactions.models.discord.base import DiscordObject, ClientObject
from interactions.models.discord.enums import InteractionPermissionTypes
from interactions.models.discord.snowflake import to_snowflake

if TYPE_CHECKING:
    from interactions import BaseContext
    from interactions import Snowflake_Type, Guild

__all__ = ("ApplicationCommandPermission",)


@attrs.define(eq=False, order=False, hash=False, kw_only=True)
class ApplicationCommandPermission(DiscordObject):
    id: "Snowflake_Type" = attrs.field(repr=False, converter=to_snowflake)
    """ID of the role user or channel"""
    type: InteractionPermissionTypes = attrs.field(repr=False, converter=InteractionPermissionTypes)
    """Type of permission (role user or channel)"""
    permission: bool = attrs.field(repr=False, default=False)
    """Whether the command is enabled for this permission"""


@attrs.define(eq=False, order=False, hash=False, kw_only=True)
class CommandPermissions(ClientObject):
    command_id: "Snowflake_Type" = attrs.field(
        repr=False,
    )
    _guild: "Guild" = attrs.field(
        repr=False,
    )

    permissions: dict["Snowflake_Type", ApplicationCommandPermission] = attrs.field(
        repr=False, factory=dict, init=False
    )

    def is_enabled(self, *object_id) -> bool:
        """
        Check if a command is enabled for given scope(s). Takes into account the permissions for the bot itself

        Args:
            *object_id: The object(s) ID to check for.

        Returns:
            Whether the command is enabled for the given scope(s).

        """
        bot_perms = self._guild.command_permissions.get(self._client.app.id)

        for obj_id in object_id:
            obj_id = to_snowflake(obj_id)
            if permission := self.permissions.get(obj_id):
                if not permission.permission:
                    return False

            if bot_perms:
                if permission := bot_perms.permissions.get(obj_id):
                    if not permission.permission:
                        return False
        return True

    def is_enabled_in_context(self, context: "BaseContext") -> bool:
        """
        Check if a command is enabled for the given context.

        Args:
            context: The context to check for.

        Returns:
            Whether the command is enabled for the given context.

        """
        everyone_role = context.guild.id
        all_channels = context.guild.id - 1  # why tf discord
        return self.is_enabled(
            context.channel.id,
            *context.member.roles,
            context.author.id,
            everyone_role,
            all_channels,
        )

    def update_permissions(self, *permissions: ApplicationCommandPermission) -> None:
        """
        Update the permissions for the command.

        Args:
            permissions: The permission to set.

        """
        self.permissions = {perm.id: perm for perm in permissions}



from asyncio import Future, iscoroutinefunction
from typing import Callable, Optional, Union, Awaitable

__all__ = ("Wait",)


class Wait:
    """Class for waiting for a future event to happen. Internally used by wait_for."""

    def __init__(
        self, event: str, checks: Optional[Union[Callable[..., bool], Callable[..., Awaitable[bool]]]], future: Future
    ) -> None:
        self.event: str = event
        self.check: Optional[Union[Callable[..., bool], Callable[..., Awaitable[bool]]]] = checks
        self.future: Future = future

    async def __call__(self, *args, **kwargs) -> bool:
        if self.future.cancelled():
            return True

        if self.check:
            try:
                if iscoroutinefunction(self.check):
                    check_result = await self.check(*args, **kwargs)
                else:
                    check_result = self.check(*args, **kwargs)
            except Exception as exc:
                self.future.set_exception(exc)
                return True
        else:
            check_result = True

        if check_result:
            self.future.set_result(*args, **kwargs)
            return True

        return False



from typing import Optional, TYPE_CHECKING

import attrs

from interactions.models.discord.timestamp import Timestamp
from interactions.models.discord.enums import EntitlementType
from interactions.models.discord.base import DiscordObject
from interactions.client.utils.attr_converters import optional as optional_c
from interactions.client.utils.attr_converters import timestamp_converter
from interactions.models.discord.snowflake import to_snowflake, to_optional_snowflake, Snowflake_Type

if TYPE_CHECKING:
    from interactions.models.discord.guild import Guild
    from interactions.models.discord.user import User

__all__ = ("Entitlement",)


@attrs.define(eq=False, order=False, hash=False, kw_only=False)
class Entitlement(DiscordObject):
    sku_id: Snowflake_Type = attrs.field(repr=False, converter=to_snowflake)
    """ID of the SKU."""
    application_id: Snowflake_Type = attrs.field(repr=False, converter=to_snowflake)
    """ID of the parent application."""
    type: EntitlementType = attrs.field(repr=False, converter=EntitlementType)
    """The type of entitlement."""
    deleted: bool = attrs.field(repr=False)
    """Whether the entitlement is deleted."""
    consumed: bool = attrs.field(repr=False, default=False)
    """For consumable items, whether or not the entitlement has been consumed"""
    subscription_id: Optional[Snowflake_Type] = attrs.field(repr=False, converter=to_optional_snowflake, default=None)
    """The ID of the subscription plan. Not present when using test entitlements."""
    starts_at: Optional[Timestamp] = attrs.field(repr=False, converter=optional_c(timestamp_converter), default=None)
    """Start date at which the entitlement is valid. Not present when using test entitlements."""
    ends_at: Optional[Timestamp] = attrs.field(repr=False, converter=optional_c(timestamp_converter), default=None)
    """Date at which the entitlement is no longer valid. Not present when using test entitlements."""
    _user_id: Optional[Snowflake_Type] = attrs.field(repr=False, converter=to_optional_snowflake, default=None)
    """The ID of the user that is granted access to the entitlement's SKU."""
    _guild_id: Optional[Snowflake_Type] = attrs.field(repr=False, converter=to_optional_snowflake, default=None)
    """The ID of the guild that is granted access to the entitlement's SKU."""

    @property
    def user(self) -> "Optional[User]":
        """The user that is granted access to the entitlement's SKU, if applicable."""
        return self.client.cache.get_user(self._user_id)

    @property
    def guild(self) -> "Optional[Guild]":
        """The guild that is granted access to the entitlement's SKU, if applicable."""
        return self.client.cache.get_guild(self._guild_id)



from typing import TYPE_CHECKING, Optional

from ..route import Route, PAYLOAD_TYPE
from interactions.models.internal.protocols import CanRequest
from interactions.models.discord.snowflake import to_optional_snowflake, to_snowflake
from interactions.client.utils.serializer import dict_filter_none

if TYPE_CHECKING:
    from interactions.models.discord.snowflake import Snowflake_Type

__all__ = ("EntitlementRequests",)


class EntitlementRequests(CanRequest):
    async def get_entitlements(
        self,
        application_id: "Snowflake_Type",
        *,
        user_id: "Optional[Snowflake_Type]" = None,
        sku_ids: "Optional[list[Snowflake_Type]]" = None,
        before: "Optional[Snowflake_Type]" = None,
        after: "Optional[Snowflake_Type]" = None,
        limit: Optional[int] = 100,
        guild_id: "Optional[Snowflake_Type]" = None,
        exclude_ended: Optional[bool] = None,
    ) -> list[dict]:
        """
        Get an application's entitlements.

        Args:
            application_id: The ID of the application.
            user_id: The ID of the user to filter entitlements by.
            sku_ids: The IDs of the SKUs to filter entitlements by.
            before: Get entitlements before this ID.
            after: Get entitlements after this ID.
            limit: The maximum number of entitlements to return. Maximum is 100.
            guild_id: The ID of the guild to filter entitlements by.
            exclude_ended: Whether to exclude ended entitlements.

        Returns:
            A dictionary containing the application's entitlements.

        """
        params: PAYLOAD_TYPE = {
            "user_id": to_optional_snowflake(user_id),
            "sku_ids": [to_snowflake(sku_id) for sku_id in sku_ids] if sku_ids else None,
            "before": to_optional_snowflake(before),
            "after": to_optional_snowflake(after),
            "limit": limit,
            "guild_id": to_optional_snowflake(guild_id),
            "exclude_ended": exclude_ended,
        }
        params = dict_filter_none(params)

        return await self.request(
            Route("GET", "/applications/{application_id}/entitlements", application_id=application_id), params=params
        )

    async def create_test_entitlement(self, payload: dict, application_id: "Snowflake_Type") -> dict:
        """
        Create a test entitlement for an application.

        Args:
            payload: The entitlement's data.
            application_id: The ID of the application.

        Returns:
            A dictionary containing the test entitlement.

        """
        return await self.request(
            Route("POST", "/applications/{application_id}/entitlements", application_id=application_id), payload=payload
        )

    async def delete_test_entitlement(self, application_id: "Snowflake_Type", entitlement_id: "Snowflake_Type") -> None:
        """
        Delete a test entitlement for an application.

        Args:
            application_id: The ID of the application.
            entitlement_id: The ID of the entitlement.

        """
        await self.request(
            Route(
                "DELETE",
                "/applications/{application_id}/entitlements/{entitlement_id}",
                application_id=application_id,
                entitlement_id=entitlement_id,
            )
        )

    async def consume_entitlement(self, application_id: "Snowflake_Type", entitlement_id: "Snowflake_Type") -> None:
        """
        For One-Time Purchase consumable SKUs, marks a given entitlement for the user as consumed.

        Args:
            application_id: The ID of the application.
            entitlement_id: The ID of the entitlement.

        """
        await self.request(
            Route(
                "POST",
                "/applications/{application_id}/entitlements/{entitlement_id}/consume",
                application_id=application_id,
                entitlement_id=entitlement_id,
            )
        )



import asyncio
import platform
import tracemalloc
from typing import ClassVar

from interactions import (
    Client,
    Extension,
    listen,
    slash_command,
    InteractionContext,
    Timestamp,
    TimestampStyles,
    Intents,
)
from interactions.client.const import __version__, __py_version__
from interactions.models.internal import checks
from .debug_application_cmd import DebugAppCMD
from .debug_exec import DebugExec
from .debug_exts import DebugExts
from .utils import get_cache_state, debug_embed

__all__ = ("DebugExtension",)


class DebugExtension(DebugExec, DebugAppCMD, DebugExts, Extension):
    class Metadata(Extension.Metadata):
        name = "Debug Extension"
        description = "Debugging utilities for interactions.py"
        version = "1.0.0"
        url = "https://github.com/interactions-py/interactions.py"
        requirements: ClassVar[list] = ["interactions>=5.0.0"]

    def __init__(self, bot: Client) -> None:
        bot.logger.info("Debug Extension is mounting!")

        super().__init__(bot)
        self.add_ext_check(checks.is_owner())

        tracemalloc.start()
        bot.logger.warning("Tracemalloc started")

        self.client.http.show_ratelimit_traceback = True
        bot.logger.warning("Ratelimit tracebacks enabled")

    async def async_start(self) -> None:
        loop = asyncio.get_running_loop()
        loop.set_debug(True)
        self.bot.logger.warning("Asyncio debug mode enabled")

    @listen()
    async def on_startup(self) -> None:
        self.bot.logger.info(f"Started {self.bot.user.tag} [{self.bot.user.id}] in Debug Mode")

        self.bot.logger.info(f"Caching System State: \n{get_cache_state(self.bot)}")

    @slash_command(
        "debug",
        sub_cmd_name="info",
        sub_cmd_description="Get basic information about the bot",
    )
    async def debug_info(self, ctx: InteractionContext) -> None:
        await ctx.defer()

        uptime = Timestamp.fromdatetime(self.bot.start_time)
        e = debug_embed("General")
        e.set_thumbnail(self.bot.user.avatar.url)
        e.add_field("Operating System", platform.platform())

        e.add_field("Version Info", f"I.py@{__version__} | Py@{__py_version__}")

        e.add_field("Start Time", f"{uptime.format(TimestampStyles.RelativeTime)}")

        if privileged_intents := [i.name for i in self.bot.intents if i in Intents.PRIVILEGED]:
            e.add_field("Privileged Intents", " | ".join(privileged_intents))

        e.add_field("Loaded Exts", ", ".join(self.bot.ext))

        e.add_field("Guilds", str(len(self.bot.guilds)))

        await ctx.send(embeds=[e])

    @debug_info.subcommand("cache", sub_cmd_description="Get information about the current cache state")
    async def cache_info(self, ctx: InteractionContext) -> None:
        await ctx.defer()
        e = debug_embed("Cache")

        e.description = f"```prolog\n{get_cache_state(self.bot)}\n```"
        await ctx.send(embeds=[e])

    @debug_info.subcommand("shutdown", sub_cmd_description="Shutdown the bot.")
    async def shutdown(self, ctx: InteractionContext) -> None:
        await ctx.send("Shutting down 😴")
        await self.bot.stop()


def setup(bot: Client) -> None:
    DebugExtension(bot)



import functools
import inspect
import re
from typing import Callable, Iterable, List, Optional, Any, Union, TYPE_CHECKING

import interactions.api.events as events
from interactions.client.const import T
from interactions.models.discord.enums import ComponentType

if TYPE_CHECKING:
    from interactions.models.discord.components import BaseComponent

__all__ = (
    "escape_mentions",
    "find",
    "find_all",
    "get",
    "get_all",
    "wrap_partial",
    "get_parameters",
    "get_event_name",
    "get_object_name",
    "maybe_coroutine",
    "nulled_boolean_get",
)

mention_reg = re.compile(r"@(everyone|here|[!&]?[0-9]{17,20})")
camel_to_snake = re.compile(r"([A-Z]+)")


def escape_mentions(content: str) -> str:
    """
    Escape mentions that could ping someone in a string.

    !!! note
        This does not escape channel mentions as they do not ping anybody

    Args:
        content: The string to escape

    Returns:
        Processed string

    """
    return mention_reg.sub("@\u200b\\1", content)


def find(predicate: Callable[[T], bool], sequence: Iterable[T]) -> Optional[T]:
    """
    Find the first element in a sequence that matches the predicate.

    ??? Hint "Example Usage:"
        ```python
        member = find(lambda m: m.name == "UserName", guild.members)
        ```
    Args:
        predicate: A callable that returns a boolean value
        sequence: A sequence to be searched

    Returns:
        A match if found, otherwise None

    """
    return next((el for el in sequence if predicate(el)), None)


def find_all(predicate: Callable[[T], bool], sequence: Iterable[T]) -> List[T]:
    """
    Find all elements in a sequence that match the predicate.

    ??? Hint "Example Usage:"
        ```python
        members = find_all(lambda m: m.name == "UserName", guild.members)
        ```
    Args:
        predicate: A callable that returns a boolean value
        sequence: A sequence to be searched

    Returns:
        A list of matches

    """
    return [el for el in sequence if predicate(el)]


def get(sequence: Iterable[T], **kwargs: Any) -> Optional[T]:
    """
    Find the first element in a sequence that matches all attrs.

    ??? Hint "Example Usage:"
        ```python
        channel = get(guild.channels, nsfw=False, category="General")
        ```

    Args:
        sequence: A sequence to be searched
        **kwargs: Keyword arguments to search the sequence for

    Returns:
        A match if found, otherwise None

    """
    if not kwargs:
        return sequence[0]

    for el in sequence:
        if any(not hasattr(el, attr) for attr in kwargs):
            continue
        if all(getattr(el, attr) == value for attr, value in kwargs.items()):
            return el
    return None


def get_all(sequence: Iterable[T], **kwargs: Any) -> List[T]:
    """
    Find all elements in a sequence that match all attrs.

    ??? Hint "Example Usage:"
        ```python
        channels = get_all(guild.channels, nsfw=False, category="General")
        ```

    Args:
        sequence: A sequence to be searched
        **kwargs: Keyword arguments to search the sequence for

    Returns:
        A list of matches

    """
    if not kwargs:
        return sequence

    matches = []
    for el in sequence:
        if any(not hasattr(el, attr) for attr in kwargs):
            continue
        if all(getattr(el, attr) == value for attr, value in kwargs.items()):
            matches.append(el)
    return matches


def wrap_partial(obj: Any, cls: Any) -> Callable:
    """
    🎁 Wraps a commands callback objects into partials.

    !!! note
        This is used internally, you shouldn't need to use this function

    Args:
        obj: The command object to process
        cls: The class to use in partials

    Returns:
        The original command object with its callback methods wrapped

    """
    if obj.callback is None or isinstance(obj.callback, functools.partial):
        return obj
    if "_no_wrap" not in getattr(obj.callback, "__name__", ""):
        obj.callback = functools.partial(obj.callback, cls)

    if getattr(obj, "error_callback", None):
        obj.error_callback = functools.partial(obj.error_callback, cls)
    if getattr(obj, "pre_run_callback", None):
        obj.pre_run_callback = functools.partial(obj.pre_run_callback, cls)
    if getattr(obj, "post_run_callback", None):
        obj.post_run_callback = functools.partial(obj.post_run_callback, cls)
    if getattr(obj, "autocomplete_callbacks", None):
        obj.autocomplete_callbacks = {k: functools.partial(v, cls) for k, v in obj.autocomplete_callbacks.items()}
    if getattr(obj, "subcommands", None):
        obj.subcommands = {k: wrap_partial(v, cls) for k, v in obj.subcommands.items()}

    return obj


def get_parameters(callback: Callable) -> dict[str, inspect.Parameter]:
    """
    Gets all the parameters of a callback.

    Args:
        callback: The callback to get the parameters of

    Returns:
        A dictionary of parameters

    """
    return {p.name: p for p in inspect.signature(callback).parameters.values()}


@functools.lru_cache(maxsize=50)
def get_event_name(event: Union[str, "events.BaseEvent"]) -> str:
    """
    Get the event name smartly from an event class or string name.

    Args:
        event: The event to parse the name of

    Returns:
        The event name

    """
    name = event

    if inspect.isclass(name) and issubclass(name, events.BaseEvent):
        name = name.__name__

    # convert CamelCase to snake_case
    name = camel_to_snake.sub(r"_\1", name).lower()
    # remove any leading underscores
    name = name.lstrip("_")
    # remove any `on_` prefixes
    name = name.removeprefix("on_")

    return name


def get_object_name(x: Any) -> str:
    """
    Gets the name of virtually any object.

    Args:
        x (Any): The object to get the name of.

    Returns:
        str: The name of the object.

    """
    try:
        return x.__name__
    except AttributeError:
        return repr(x) if hasattr(x, "__origin__") else x.__class__.__name__


async def maybe_coroutine(func: Callable, *args, **kwargs) -> Any:
    """Allows running either a coroutine or a function."""
    if inspect.iscoroutinefunction(func):
        return await func(*args, **kwargs)
    return func(*args, **kwargs)


def disable_components(*components: "BaseComponent") -> list["BaseComponent"]:
    """Disables all components in a list of components."""
    for component in components:
        if component.type == ComponentType.ACTION_ROW:
            disable_components(*component.components)
        else:
            component.disabled = True
    return list(components)


def nulled_boolean_get(data: dict[str, Any], key: str) -> bool:
    """
    Gets a boolean value from a dictionary, but treats None as True.

    Args:
        data: The dictionary to get the value from
        key: The key to get the value from

    Returns:
        The boolean value of the key

    """
    # discord tags are weird, when they are None they are True, when they are True they are True and when they are False they are False
    if key in data:
        return True if data[key] is None else bool(data[key])
    return False



from .annotations import (
    slash_attachment_option,
    slash_bool_option,
    slash_channel_option,
    slash_float_option,
    slash_int_option,
    slash_mentionable_option,
    slash_role_option,
    slash_str_option,
    slash_user_option,
)
from .callback import CallbackObject
from .active_voice_state import ActiveVoiceState
from .auto_defer import AutoDefer  # purposely out of order to make sure auto_defer comes out as the deco
from .application_commands import (
    application_commands_to_dict,
    auto_defer,
    CallbackType,
    component_callback,
    ComponentCommand,
    contexts,
    context_menu,
    user_context_menu,
    message_context_menu,
    ContextMenu,
    global_autocomplete,
    GlobalAutoComplete,
    integration_types,
    InteractionCommand,
    LocalisedDesc,
    LocalisedName,
    LocalizedDesc,
    LocalizedName,
    modal_callback,
    ModalCommand,
    OptionType,
    slash_command,
    slash_default_member_permission,
    slash_option,
    SlashCommand,
    SlashCommandChoice,
    SlashCommandOption,
    SlashCommandParameter,
    subcommand,
    sync_needed,
)
from .checks import dm_only, guild_only, has_any_role, has_id, has_role, is_owner
from .command import BaseCommand, check, cooldown, max_concurrency
from .context import (
    AutocompleteContext,
    BaseContext,
    BaseInteractionContext,
    ComponentContext,
    ContextMenuContext,
    InteractionContext,
    ModalContext,
    Resolved,
    SlashContext,
)
from .converters import (
    BaseChannelConverter,
    ChannelConverter,
    ConsumeRest,
    CustomEmojiConverter,
    DMChannelConverter,
    DMConverter,
    DMGroupConverter,
    Greedy,
    GuildCategoryConverter,
    GuildChannelConverter,
    GuildConverter,
    GuildNewsConverter,
    GuildNewsThreadConverter,
    GuildPrivateThreadConverter,
    GuildPublicThreadConverter,
    GuildStageVoiceConverter,
    GuildTextConverter,
    GuildVoiceConverter,
    IDConverter,
    MemberConverter,
    MessageableChannelConverter,
    MessageConverter,
    MODEL_TO_CONVERTER,
    NoArgumentConverter,
    PartialEmojiConverter,
    RoleConverter,
    SnowflakeConverter,
    ThreadChannelConverter,
    UserConverter,
    VoiceChannelConverter,
)
from .cooldowns import (
    Buckets,
    Cooldown,
    CooldownSystem,
    MaxConcurrency,
    SlidingWindowSystem,
    ExponentialBackoffSystem,
    LeakyBucketSystem,
    TokenBucketSystem,
)
from .listener import listen, Listener
from .protocols import Converter
from .extension import Extension
from .wait import Wait
from .tasks import BaseTrigger, DateTrigger, IntervalTrigger, OrTrigger, Task, TimeTrigger, CronTrigger

__all__ = (
    "ActiveVoiceState",
    "application_commands_to_dict",
    "auto_defer",
    "AutocompleteContext",
    "AutoDefer",
    "BaseChannelConverter",
    "BaseCommand",
    "BaseContext",
    "BaseInteractionContext",
    "BaseTrigger",
    "Buckets",
    "CallbackObject",
    "CallbackType",
    "ChannelConverter",
    "check",
    "component_callback",
    "ComponentCommand",
    "ComponentContext",
    "contexts",
    "context_menu",
    "user_context_menu",
    "message_context_menu",
    "ConsumeRest",
    "ContextMenu",
    "ContextMenuContext",
    "Converter",
    "cooldown",
    "Cooldown",
    "CooldownSystem",
    "CronTrigger",
    "SlidingWindowSystem",
    "ExponentialBackoffSystem",
    "LeakyBucketSystem",
    "TokenBucketSystem",
    "CustomEmojiConverter",
    "DateTrigger",
    "dm_only",
    "DMChannelConverter",
    "DMConverter",
    "DMGroupConverter",
    "Extension",
    "global_autocomplete",
    "GlobalAutoComplete",
    "Greedy",
    "guild_only",
    "GuildCategoryConverter",
    "GuildChannelConverter",
    "GuildConverter",
    "GuildNewsConverter",
    "GuildNewsThreadConverter",
    "GuildPrivateThreadConverter",
    "GuildPublicThreadConverter",
    "GuildStageVoiceConverter",
    "GuildTextConverter",
    "GuildVoiceConverter",
    "has_any_role",
    "has_id",
    "has_role",
    "IDConverter",
    "integration_types",
    "InteractionCommand",
    "InteractionContext",
    "IntervalTrigger",
    "is_owner",
    "listen",
    "Listener",
    "LocalisedDesc",
    "LocalisedName",
    "LocalizedDesc",
    "LocalizedName",
    "max_concurrency",
    "MaxConcurrency",
    "MemberConverter",
    "MessageableChannelConverter",
    "MessageConverter",
    "modal_callback",
    "ModalCommand",
    "ModalContext",
    "MODEL_TO_CONVERTER",
    "NoArgumentConverter",
    "OptionType",
    "OrTrigger",
    "PartialEmojiConverter",
    "Resolved",
    "RoleConverter",
    "slash_attachment_option",
    "slash_bool_option",
    "slash_channel_option",
    "slash_command",
    "slash_default_member_permission",
    "slash_float_option",
    "slash_int_option",
    "slash_mentionable_option",
    "slash_option",
    "slash_role_option",
    "slash_str_option",
    "slash_user_option",
    "SlashCommand",
    "SlashCommandChoice",
    "SlashCommandOption",
    "SlashCommandParameter",
    "SlashContext",
    "SnowflakeConverter",
    "subcommand",
    "sync_needed",
    "Task",
    "ThreadChannelConverter",
    "TimeTrigger",
    "UserConverter",
    "VoiceChannelConverter",
    "Wait",
)
